// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2003 Martin Schweiger
//                   All rights reserved
//
// Atlantis.cpp
// Reference implementation of Atlantis (Space Shuttle) vessel
// class module
//
// RMS, grappling and MMU capabilities by Robert Conley
// ==============================================================

#define STRICT
#include "Atlantis.h"
#include "resource.h"
#include <stdio.h>
#include <fstream.h>

// ==============================================================
// Global (class-wide) parameters

struct {
	HINSTANCE hDLL;
} g_Param;

// ==============================================================
// Local prototypes

BOOL CALLBACK RMS_DlgProc (HWND, UINT, WPARAM, LPARAM);
extern void GetSRB_State (double met, double &thrust_level, double &prop_level);


// ==============================================================
// Airfoil coefficient functions
// Return lift, moment and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// ==============================================================

// 1. vertical lift component (wings and body)

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38,  -0.5,  -0.5, -0.02, 0.6355,    0.63,   0.46, 0.28, 0.13, 0.0, -0.16, -0.26, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = {  0,    0,   0,   0,    0,   0, 0,     0,    0,0.002,0.004, 0.0025,0.0012,      0,-0.0012,-0.0007,    0,    0,   0,     0,     0,     0,     0,    0,   0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = CM[idx] + (CM[idx+1]-CM[idx])*d;
	*cd = 0.06 + oapiGetInducedDrag (*cl, 2.266, 0.6);
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.2, 0.3, 0.2, 0, -0.2, -0.3, -0.2, 0, 0.2, 0.3, 0.2, 0, -0.2, -0.3, -0.2, 0};

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.02 + oapiGetInducedDrag (*cl, 1.5, 0.6);
}

// ==============================================================
// Specialised vessel class Atlantis
// ==============================================================

// Constructor
Atlantis::Atlantis (OBJHANDLE hObj, int fmodel)
: VESSEL(hObj, fmodel)
{
	status          = 0;
	gear_status     = DOOR_CLOSED;
	gear_proc       = 0.0;
	door_status     = DOOR_CLOSED;
	door_proc       = ldoor_drag = rdoor_drag = 0.0;
	kubd_status     = DOOR_CLOSED;
	kubd_proc       = 0.0;
	spdb_status     = DOOR_CLOSED;
	spdb_proc       = 0.0;

	vis             = NULL;
	
	reset_mmu       = false;
	reset_sat       = false;
	jettison_time   = 0.0;
	render_cockpit  = false;

	// propellant resources
	ph_oms          = NULL;
	ph_tank         = NULL;
	ph_srb          = NULL;
	thg_main        = NULL;
	thg_srb         = NULL;

	// preload meshes
	hOrbiterMesh = oapiLoadMeshGlobal ("Atlantis");
	hTankMesh    = oapiLoadMeshGlobal ("Atlantis_tank");
	hSRBMesh     = oapiLoadMeshGlobal ("Atlantis_srb");


	DefineAnimations();
	center_arm      = false;
	arm_moved       = false;
	bManualSeparate = false;
	ofs_sts_sat     = _V(0,0,0);      
	do_eva          = false;
	do_plat         = false;
	do_cargostatic  = false;
	vis             = NULL;
	sat_attach      = NULL;
	rms_attach      = NULL;
	cargo_static_ofs   =_V(0,0,0);

	// default arm status: stowed
	arm_sy = 0.5;
	arm_sp = 0.0;
	arm_ep = 0.0;
	arm_wp = 0.5;
	arm_wy = 0.5;
	arm_wr = 0.5;
	arm_tip[0] = _V(-2.26,1.71,-6.5);
	arm_tip[1] = _V(-2.26,1.71,-7.5);
	arm_tip[2] = _V(-2.26,2.71,-6.5);
}

Atlantis::~Atlantis ()
{
	int i;
	for (i = 0; i < 6; i++) delete rms_anim[i];
}

// Set vessel class capabilities from config file
void Atlantis::SetClassCaps (FILEHANDLE cfg)
{
	if (!oapiReadItem_bool (cfg, "RenderCockpit", render_cockpit))
		render_cockpit = false;
}

// Read status from scenario file
void Atlantis::LoadState (FILEHANDLE scn, void *vs)
{
    char *line;
	double met = 0.0; // mission elapsed time
	double srbtime = 0.0;
	double sts_sat_x = 0.0;
	double sts_sat_y = 0.0;
	double sts_sat_z = 0.0;
	spdb_status = DOOR_CLOSED; spdb_proc = 0.0;

	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		} else if (!strnicmp (line, "MET", 3)) {
			sscanf (line+3, "%lf", &met);
		} else if (!strnicmp (line, "CARGODOOR", 9)) {
			sscanf (line+9, "%d%lf", &door_status, &door_proc);
		} else if (!strnicmp (line, "GEAR", 4)) {
			sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "KUBAND", 6)) {
			sscanf (line+6, "%d%lf", &kubd_status, &kubd_proc);
		} else if (!strnicmp (line, "SPEEDBRAKE", 10)) {
			sscanf (line+10, "%d%lf", &spdb_status, &spdb_proc);
		} else if (!strnicmp (line, "SRB_IGNITION_TIME", 17)) {
			sscanf (line+17, "%lf", &srbtime);
		} else if (!strnicmp (line, "SAT_OFS_X", 9)) {
			sscanf (line+9, "%lf", &sts_sat_x);
		} else if (!strnicmp (line, "SAT_OFS_Y", 9)) {
			sscanf (line+9, "%lf", &sts_sat_y);
		} else if (!strnicmp (line, "SAT_OFS_Z", 9)) {
			sscanf (line+9, "%lf", &sts_sat_z);
		} else if (!strnicmp (line, "CARGO_STATIC_MESH", 17)) {
			sscanf (line+17, "%s", cargo_static_mesh_name);
			do_cargostatic = true;
		} else if (!strnicmp (line, "CARGO_STATIC_OFS", 16)) {
			sscanf (line+16, "%lf%lf%lf", &cargo_static_ofs.x, &cargo_static_ofs.y, &cargo_static_ofs.z);
		} else if (!strnicmp (line, "ARM_STATUS", 10)) {
			sscanf (line+10, "%lf%lf%lf%lf%lf%lf", &arm_sy, &arm_sp, &arm_ep, &arm_wp, &arm_wy, &arm_wr);
        } else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
    }
	ofs_sts_sat.x=sts_sat_x;
	ofs_sts_sat.y=sts_sat_y;
	ofs_sts_sat.z=sts_sat_z;

	switch (status) {
	case 0:
		SetLaunchConfiguration();
		break;
	case 1:
		SetPostLaunchConfiguration (met);
		break;
	case 2:
		SetOrbiterTankConfiguration();
		break;
	case 3:
		SetOrbiterConfiguration();
		break;
	}

	UpdateMesh ();
}

// Write status to scenario file
void Atlantis::SaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);

	if (status == 1)
		oapiWriteScenario_float (scn, "MET", oapiGetSimTime()-t0);

	sprintf (cbuf, "%d %0.4f", door_status, door_proc);
	oapiWriteScenario_string (scn, "CARGODOOR", cbuf);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);

	sprintf (cbuf, "%d %0.4f", kubd_status, kubd_proc);
	oapiWriteScenario_string (scn, "KUBAND", cbuf);

	if (spdb_status != DOOR_CLOSED) {
		sprintf (cbuf, "%d %0.4f", spdb_status, spdb_proc);
		oapiWriteScenario_string (scn, "SPEEDBRAKE", cbuf);
	}

	sprintf (cbuf, "%0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", arm_sy, arm_sp, arm_ep, arm_wp, arm_wy, arm_wr);
	oapiWriteScenario_string (scn, "ARM_STATUS", cbuf);

	oapiWriteScenario_float (scn, "SAT_OFS_X", ofs_sts_sat.x);
	oapiWriteScenario_float (scn, "SAT_OFS_Y", ofs_sts_sat.y);
	oapiWriteScenario_float (scn, "SAT_OFS_Z", ofs_sts_sat.z);

	if (do_cargostatic) {
		oapiWriteScenario_string (scn, "CARGO_STATIC_MESH", cargo_static_mesh_name);
		if (cargo_static_ofs.x || cargo_static_ofs.y || cargo_static_ofs.z)
			oapiWriteScenario_vec (scn, "CARGO_STATIC_OFS", cargo_static_ofs);
	}
}

// launch configuration setup (orbiter + tank + SRB)
void Atlantis::SetLaunchConfiguration (void)
{
	int i;
	extern PARTICLESTREAMSPEC srb_contrail, srb_exhaust;

	// *********************** physical parameters *********************************

	SetSize (30.0);
	SetEmptyMass (ORBITER_EMPTY_MASS + TANK_EMPTY_MASS + 2*SRB_EMPTY_MASS);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(687.4,849.5,189.4));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(179.1,176.8,29.3));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_LAUNCH_ORBITER, _V(0,1,0), _V(0,0,-1));
	SetTouchdownPoints (_V(0,-10,-53.3), _V(-7,7,-53.3), _V(7,7,-53.3));

	// ************************* propellant specs **********************************

	if (!ph_oms)  ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	if (!ph_tank) ph_tank = CreatePropellantResource (TANK_MAX_PROPELLANT_MASS);    // main tank
	if (!ph_srb)  ph_srb  = CreatePropellantResource (SRB_MAX_PROPELLANT_MASS*2.0); // SRB's
	SetDefaultPropellantResource (ph_tank); // display main tank level in generic HUD

	// *********************** thruster definitions ********************************

	// The main and SRB thrusters are defined so as to minimise the angular momentum
	// when engaged at maximum thrust. Dynamic gimbaling is not required until the SRBs
	// start to lose power since the centre of gravity is assumed static.
	// However the resulting linear force vector has a component in +y ("up") direction

	ClearThrusterDefinitions();

	// orbiter main thrusters
	th_main[0] = CreateThruster (OFS_LAUNCH_ORBITER+_V(-1.6,-0.2,-16.0), _V( 0.04994,0.0,0.99875), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_main[1] = CreateThruster (OFS_LAUNCH_ORBITER+_V( 1.6,-0.2,-16.0), _V(-0.04994,0.0,0.99875), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_main[2] = CreateThruster (OFS_LAUNCH_ORBITER+_V( 0.0, 3.2,-15.5), _V( 0.0,-0.13,1), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	thg_main = CreateThrusterGroup (th_main, 3, THGROUP_MAIN);
	SURFHANDLE tex_main = oapiRegisterExhaustTexture ("Exhaust_atsme");
	for (i = 0; i < 3; i++) AddExhaust (th_main[i], 30.0, 2.0, tex_main);

	// SRBs
	th_srb[0] = CreateThruster (OFS_LAUNCH_RIGHTSRB+_V(0.0,0.0,-21.8), _V(0,0.023643,0.999720), SRB_THRUST, ph_srb, SRB_ISP0, SRB_ISP1);
	th_srb[1] = CreateThruster (OFS_LAUNCH_LEFTSRB +_V(0.0,0.0,-21.8), _V(0,0.023643,0.999720), SRB_THRUST, ph_srb, SRB_ISP0, SRB_ISP1);
	thg_srb = CreateThrusterGroup (th_srb, 2, THGROUP_USER);
	SURFHANDLE tex = oapiRegisterExhaustTexture ("Exhaust2");
	srb_exhaust.tex = oapiRegisterParticleTexture ("Contrail2");
	for (i = 0; i < 2; i++) AddExhaust (th_srb[i], 16.0, 2.0, tex);
	AddExhaustStream (th_srb[0], OFS_LAUNCH_RIGHTSRB+_V(0,0,-30), &srb_contrail);
	AddExhaustStream (th_srb[1], OFS_LAUNCH_LEFTSRB+_V(0,0,-30), &srb_contrail);
	AddExhaustStream (th_srb[0], OFS_LAUNCH_RIGHTSRB+_V(0,0,-25), &srb_exhaust);
	AddExhaustStream (th_srb[1], OFS_LAUNCH_LEFTSRB+_V(0,0,-25), &srb_exhaust);

	// attitude - this is temporary
	// attitude adjustment during launch phase should really be done via SRB gimbaling
	CreateAttControls_Launch();

	// ************************* aerodynamics **************************************

	ClearVariableDragElements ();
	CreateVariableDragElement (&spdb_proc, 5, OFS_LAUNCH_ORBITER+_V(0, 7.5, -14)); // speedbrake drag
	CreateVariableDragElement (&gear_proc, 2, OFS_LAUNCH_ORBITER+_V(0,-3,0));      // landing gear drag
	CreateVariableDragElement (&rdoor_drag, 7, OFS_LAUNCH_ORBITER+_V(2.9,0,10));   // right cargo door drag
	CreateVariableDragElement (&ldoor_drag, 7, OFS_LAUNCH_ORBITER+_V(-2.9,0,10));  // right cargo door drag

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_LAUNCH_ORBITER);
	AddTankVisual    (OFS_LAUNCH_TANK);
	AddSRBVisual     (OFS_LAUNCH_RIGHTSRB);
	AddSRBVisual     (OFS_LAUNCH_LEFTSRB);

	status = 0;
}

void Atlantis::SetPostLaunchConfiguration (double met)
{
	SetLaunchConfiguration();
	t0 = -met; // reference time (liftoff)
	status = 1;
}

// Configuration after booster separation (orbiter + tank)
void Atlantis::SetOrbiterTankConfiguration (void)
{
	int i;
	VECTOR3 ofs;
	
	// *********************** physical parameters *********************************

	SetSize (28.8);
	SetEmptyMass (ORBITER_EMPTY_MASS + TANK_EMPTY_MASS);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(646.1,603.0,141.5));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(173.3,161.0,24.0));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_WITHTANK_ORBITER, _V(0,1,0), _V(0,0,-1));
	SetTouchdownPoints (_V(0,-5,30), _V(-10,-10,-30), _V(10,0,-30));

	// ************************* propellant specs **********************************

	if (!ph_oms)  ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	if (!ph_tank) ph_tank = CreatePropellantResource (TANK_MAX_PROPELLANT_MASS);    // main tank
	SetDefaultPropellantResource (ph_tank); // display main tank level in generic HUD

	// *********************** thruster definitions ********************************

	// Orbiter main engines
	// The thruster directions are adjusted so as to generate no angular moment
	// i.e. sum_{i=1}^3 thrusterpos_i x thrusterdir_i = 0
	// This assumes that all 3 main engines generate the same amount of thrust at
	// all times

	ofs = OFS_WITHTANK_ORBITER;
	if (thg_main) { // main engines already defined - just modify parameters
		SetThrusterRef (th_main[0], ofs+_V(-1.6,-0.2,-16.0));
		SetThrusterDir (th_main[0], _V( 0.0624,-0.1789,0.9819));
		SetThrusterRef (th_main[1], ofs+_V( 1.6,-0.2,-16.0));
		SetThrusterDir (th_main[1], _V(-0.0624,-0.1789,0.9819));
		SetThrusterRef (th_main[2], ofs+_V( 0.0, 3.2,-15.5));
		SetThrusterDir (th_main[2], _V( 0.0,   -0.308046,0.951372));
	} else {        // create main engines
		th_main[0] = CreateThruster (ofs+_V(-1.6,-0.2,-16.0), _V( 0.0624,-0.1789,0.9819), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		th_main[1] = CreateThruster (ofs+_V( 1.6,-0.2,-16.0), _V(-0.0624,-0.1789,0.9819), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		th_main[2] = CreateThruster (ofs+_V( 0.0, 3.2,-15.5), _V( 0.0,   -0.308046,0.951372), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		thg_main = CreateThrusterGroup (th_main, 3, THGROUP_MAIN);
		SURFHANDLE tex_main = oapiRegisterExhaustTexture ("Exhaust_atsme");
		for (i = 0; i < 3; i++) AddExhaust (th_main[i], 30.0, 2.0, tex_main);
	}
	if (!ThrusterGroupDefined (THGROUP_ATT_PITCHUP))
		CreateAttControls_Launch();

	// ************************* aerodynamics **************************************

	ClearVariableDragElements ();
	CreateVariableDragElement (&spdb_proc, 5, ofs+_V(0, 7.5, -14)); // speedbrake drag
	CreateVariableDragElement (&gear_proc, 2, ofs+_V(0,-3,0));      // landing gear drag
	CreateVariableDragElement (&rdoor_drag, 7, ofs+_V(2.9,0,10));   // right cargo door drag
	CreateVariableDragElement (&ldoor_drag, 7, ofs+_V(-2.9,0,10));  // right cargo door drag

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_WITHTANK_ORBITER);
	AddTankVisual    (OFS_WITHTANK_TANK);

	status = 2;
}

// Configuration after tank separation (orbiter only)
void Atlantis::SetOrbiterConfiguration (void)
{
	int i;

	// *********************** physical parameters *********************************

	SetSize (19.6);
	SetEmptyMass (ORBITER_EMPTY_MASS);
	SetPMI (_V(78.2,82.1,10.7));
	SetCW (ORBITER_CW[0], ORBITER_CW[1], ORBITER_CW[2], ORBITER_CW[3]);
	SetCrossSections (ORBITER_CS);
	SetGearParameters (gear_proc);

	if (GetFlightModel() >= 1) { // realistic flight model
	//	SetPitchMomentScale (1e-5);
	//	SetBankMomentScale (2e-5);
		if (gear_status == DOOR_OPEN) {
			SetCW (ORBITER_CW[0]+ORBITER_CW_GEAR[0],
				   ORBITER_CW[1]+ORBITER_CW_GEAR[1],
				   ORBITER_CW[2]+ORBITER_CW_GEAR[2],
				   ORBITER_CW[3]+ORBITER_CW_GEAR[3]);
			SetCrossSections (ORBITER_CS+ORBITER_CS_GEAR);
		}
	}

	// ************************* aerodynamics **************************************

	SetRotDrag (_V(0.43,0.43,0.29)); // angular drag

	CreateAirfoil (LIFT_VERTICAL,   _V(0,0,-0.5), VLiftCoeff, 20, 270, 2.266);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 20,  50, 1.5);

	CreateControlSurface (AIRCTRL_ELEVATOR, 5.0, 1.5, _V( 0, 0,  -15), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_RUDDER,   2.0, 1.5, _V( 0, 3,  -16), AIRCTRL_AXIS_YPOS);
	CreateControlSurface (AIRCTRL_AILERON,  3.0, 1.5, _V( 7,-0.5,-15), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_AILERON,  3.0, 1.5, _V(-7,-0.5,-15), AIRCTRL_AXIS_XNEG);

	ClearVariableDragElements ();
	CreateVariableDragElement (&spdb_proc, 5, _V(0, 7.5, -14)); // speedbrake drag
	CreateVariableDragElement (&gear_proc, 2, _V(0,-3,0));      // landing gear drag
	CreateVariableDragElement (&rdoor_drag, 7, _V(2.9,0,10));   // right cargo door drag
	CreateVariableDragElement (&ldoor_drag, 7, _V(-2.9,0,10));  // right cargo door drag

	SetADCtrlMode (7);

	// ************************* particle streams **********************************

	PARTICLESTREAMSPEC rps = {
		0, 20, 20, 0, 0.03, 0.5, 100, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1, PARTICLESTREAMSPEC::ATM_PLIN, 6e7, 12e7
	};
	AddReentryStream (&rps);

	SetDockParams (ORBITER_DOCKPOS, _V(0,1,0), _V(0,0,-1));
	// for docking attachment in payload bay

	// ************************* propellant specs **********************************

	if (!ph_oms) ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	SetDefaultPropellantResource (ph_oms); // display OMS tank level in generic HUD

	// *********************** thruster definitions ********************************

	// OMS (Orbital Manouevering System)
	th_main[0] = CreateThruster (_V(-2.7,3.7,-13.8), _V( 0.03,-0.25885,0.96545), ORBITER_OMS_THRUST, ph_oms, ORBITER_OMS_ISP0, ORBITER_OMS_ISP1);
	th_main[1] = CreateThruster (_V( 2.7,3.7,-13.8), _V(-0.03,-0.25885,0.96545), ORBITER_OMS_THRUST, ph_oms, ORBITER_OMS_ISP0, ORBITER_OMS_ISP1);
	thg_main = CreateThrusterGroup (th_main, 2, THGROUP_MAIN);
	for (i = 0; i < 2; i++) AddExhaust (th_main[i], 4.0, 0.5);

	// RCS (Reaction Control System)
	CreateAttControls_RCS ();

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_ZERO);

	status = 3;
}

void Atlantis::CreateAttControls_Launch()
{
	// we should only require pitch and bank, but yaw is also defined to make KILLROT work
	// no linear attitude modes are defined at this stage
	THRUSTER_HANDLE th_rot[2];
	th_rot[0] = CreateThruster (_V(0,0, 25), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-25), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);
	th_rot[0] = CreateThruster (_V(0,0, 25), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-25), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);
	th_rot[0] = CreateThruster (_V( 8,0,0), _V(0, 1,0), MAX_ROLL_SRB, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(-8,0,0), _V(0,-1,0), MAX_ROLL_SRB, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);
	th_rot[0] = CreateThruster (_V( 8,0,0), _V(0,-1,0), MAX_ROLL_SRB, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(-8,0,0), _V(0, 1,0), MAX_ROLL_SRB, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V( 1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V(-1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V(-1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V( 1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);
}

void Atlantis::CreateAttControls_RCS()
{
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("Exhaust_atrcs");
	const double eh = 6.0;             // exhaust length scale
	const double ew1 = 0.4, ew2 = 0.8; // exhaust width scales

	// set of attitude thrusters (idealised). The arrangement is such that no angular
	// momentum is created in linear mode, and no linear momentum is created in rotational mode.
	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];
	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15.5), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-15.5), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 15.5), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15.5), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_UP);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_DOWN);

	AddExhaust (th_att_rot[0], eh, ew1, _V( 1.80, 0, 19.0), _V( 0.174,-0.985,0), tex_rcs);//F2D
	AddExhaust (th_att_rot[0], eh, ew1, _V( 1.80, 0, 18.6), _V( 0.174,-0.985,0), tex_rcs);//F4D
	AddExhaust (th_att_rot[0], eh, ew1, _V(-1.75, 0, 19.2), _V( -0.174,-0.985,0), tex_rcs);//F1D
	AddExhaust (th_att_rot[0], eh, ew1, _V(-1.75, 0, 18.8), _V( -0.174,-0.985,0), tex_rcs);//F3D

	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-12.7), _V(0, 1,0), tex_rcs);//L4U
	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-13.0), _V(0, 1,0), tex_rcs);//L2U
	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-13.3), _V(0, 1,0), tex_rcs);//L1U

	AddExhaust (th_att_rot[1], eh, ew1, _V( 4.0, 3.3,-12.7), _V(0, 1,0), tex_rcs);//R4U
	AddExhaust (th_att_rot[1], eh, ew1, _V( 4.0, 3.3,-13.0), _V(0, 1,0), tex_rcs);//R2U
	AddExhaust (th_att_rot[1], eh, ew1, _V( 4.0, 3.3,-13.3), _V(0, 1,0), tex_rcs);//R1U

	AddExhaust (th_att_rot[2], eh, ew1, _V(-0.3, 1.15,18.5  ), _V(0, 1,0), tex_rcs);//F1U	
	AddExhaust (th_att_rot[2], eh, ew1, _V( 0.1,   1.2 ,18.5  ), _V(0, 1,0), tex_rcs);//F3U
	AddExhaust (th_att_rot[2], eh, ew1, _V( 0.5, 1.15,18.5  ), _V(0, 1,0), tex_rcs);//F2U

	AddExhaust (th_att_rot[3], eh, ew1, _V(-4.0, 1.9,-12.7), _V(0,-1,0), tex_rcs);//L4D
	AddExhaust (th_att_rot[3], eh, ew1, _V(-4.0, 1.9,-13.0), _V(0,-1,0), tex_rcs);//L2D
	AddExhaust (th_att_rot[3], eh, ew1, _V(-4.0, 1.9,-13.3), _V(0,-1,0), tex_rcs);//L3D

	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 1.9,-12.7), _V(0,-1,0), tex_rcs);//R4D
	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 1.9,-13.0), _V(0,-1,0), tex_rcs);//R2D
	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 1.9,-13.3), _V(0,-1,0), tex_rcs);//R3D

	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15.5), _V(-1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-15.5), _V( 1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 15.5), _V( 1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15.5), _V(-1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWRIGHT);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_LEFT);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_RIGHT);

	AddExhaust (th_att_rot[0], eh, ew2, _V( 2.0,-0.15, 18.4), _V( 1,0,0), tex_rcs);//F4R
	AddExhaust (th_att_rot[0], eh, ew2, _V( 2.0,0.15, 18.4), _V( 1,0,0), tex_rcs);//F2R
	AddExhaust (th_att_rot[2], eh, ew2, _V(-1.7,0.22, 18.5), _V(-1,0,0), tex_rcs);//F1L
	AddExhaust (th_att_rot[2], eh, ew2, _V(-1.7,-0.11, 18.5), _V(-1,0,0), tex_rcs);//F3L

	AddExhaust (th_att_rot[1], eh, ew2, _V(-4.3,   2.45,-12.7), _V(-1,0,0), tex_rcs);//L4L
	AddExhaust (th_att_rot[1], eh, ew2, _V(-4.3,   2.45,-13.05), _V(-1,0,0), tex_rcs);//L4L
	AddExhaust (th_att_rot[1], eh, ew2, _V(-4.3,   2.45,-13.4), _V(-1,0,0), tex_rcs);//L4L
	AddExhaust (th_att_rot[1], eh, ew2, _V(-4.3,   2.45,-13.65), _V(-1,0,0), tex_rcs);//L4L

	AddExhaust (th_att_rot[3], eh, ew2, _V( 4.5,   2.45,-12.7), _V( 1,0,0), tex_rcs);//R4R
	AddExhaust (th_att_rot[3], eh, ew2, _V( 4.5,   2.45,-13.05), _V( 1,0,0), tex_rcs);//R2R
	AddExhaust (th_att_rot[3], eh, ew2, _V( 4.5,   2.45,-13.4), _V( 1,0,0), tex_rcs);//R3R
	AddExhaust (th_att_rot[3], eh, ew2, _V( 4.5,   2.45,-13.65), _V( 1,0,0), tex_rcs);//R1R

	th_att_rot[0] = CreateThruster (_V( 2.7,0,0), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = CreateThruster (_V(-2.7,0,0), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = CreateThruster (_V(-2.7,0,0), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = CreateThruster (_V( 2.7,0,0), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);

	AddExhaust (th_att_rot[0], eh, ew1, _V( 1.80, 0, 19.0), _V( 0.174,-0.985,0), tex_rcs);//F2D
	AddExhaust (th_att_rot[0], eh, ew1, _V( 1.80, 0, 18.6), _V( 0.174,-0.985,0), tex_rcs);//F4D
	AddExhaust (th_att_rot[2], eh, ew1, _V(-1.75, 0, 19.2), _V( -0.174,-0.985,0), tex_rcs);//F1D
	AddExhaust (th_att_rot[2], eh, ew1, _V(-1.75, 0, 18.8), _V( -0.174,-0.985,0), tex_rcs);//F3D

	AddExhaust (th_att_rot[0], eh, ew1, _V( 4.0, 1.9,-12.7), _V(0,-1,0), tex_rcs);//R4D
	AddExhaust (th_att_rot[0], eh, ew1, _V( 4.0, 1.9,-13.0), _V(0,-1,0), tex_rcs);//R2D
	AddExhaust (th_att_rot[0], eh, ew1, _V( 4.0, 1.9,-13.3), _V(0,-1,0), tex_rcs);//R3D

	AddExhaust (th_att_rot[2], eh, ew1, _V(-4.0, 1.9,-12.7), _V(0,-1,0), tex_rcs);//L4D
	AddExhaust (th_att_rot[2], eh, ew1, _V(-4.0, 1.9,-13.0), _V(0,-1,0), tex_rcs);//L2D
	AddExhaust (th_att_rot[2], eh, ew1, _V(-4.0, 1.9,-13.3), _V(0,-1,0), tex_rcs);//L3D

	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 3.3,-12.7), _V(0, 1,0), tex_rcs);//R4U
	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 3.3,-13.0), _V(0, 1,0), tex_rcs);//R2U
	AddExhaust (th_att_rot[3], eh, ew1, _V( 4.0, 3.3,-13.3), _V(0, 1,0), tex_rcs);//R1U

	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-12.7), _V(0, 1,0), tex_rcs);//L4U
	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-13.0), _V(0, 1,0), tex_rcs);//L2U
	AddExhaust (th_att_rot[1], eh, ew1, _V(-4.0, 3.3,-13.3), _V(0, 1,0), tex_rcs);//L1U

	th_att_lin[0] = CreateThruster (_V(0,0,-16), _V(0,0, 1), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_lin[1] = CreateThruster (_V(0,0, 16), _V(0,0,-1), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_lin,   1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_att_lin+1, 1, THGROUP_ATT_BACK);

	AddExhaust (th_att_lin[0], eh, ew1, _V(-3.50,2.7,-14  ), _V(0,0,-1), tex_rcs);//L1A
	AddExhaust (th_att_lin[0], eh, ew1, _V(-3.95,2.7,-14  ), _V(0,0,-1), tex_rcs);//L3A
	AddExhaust (th_att_lin[0], eh, ew1, _V( 3.65,2.75,-14  ), _V(0,0,-1), tex_rcs);//R1A
	AddExhaust (th_att_lin[0], eh, ew1, _V( 4.05,2.75,-14  ), _V(0,0,-1), tex_rcs);//R3A

	AddExhaust (th_att_lin[1], eh, ew1, _V( 0,  0.6, 20), _V(0,0, 1), tex_rcs);//F2F
	AddExhaust (th_att_lin[1], eh, ew1, _V(-0.4,0.55,20), _V(0,0, 1), tex_rcs);//F1F
	AddExhaust (th_att_lin[1], eh, ew1, _V( 0.4,0.55,20), _V(0,0, 1), tex_rcs);//F2F
}

// Define animation sequences for moving parts
void Atlantis::DefineAnimations (void)
{
	// ***** 1. Cargo door animation *****

	static UINT RCargoDoorGrp[6] = {29,30,31,32,33,34};
	static MGROUP_ROTATE RCargoDoor (0, RCargoDoorGrp, 6,
		_V(2.9, 1.3, 0), _V(0,0,1), (float)(-175.5*RAD));
	static UINT LCargoDoorGrp[6] = {27,28,35,36,37,38};
	static MGROUP_ROTATE LCargoDoor (0, LCargoDoorGrp, 6,
		_V(-2.81, 1.3, 0), _V(0,0,1), (float)(175.5*RAD));
	static UINT RRadiatorGrp[2] = {29,30};
	static MGROUP_ROTATE RRadiator (0, RRadiatorGrp, 2,
		_V(2.9, 1.3, 0), _V(0,0,1), (float)(35.5*RAD));
	static UINT LRadiatorGrp[2] = {27,28};
	static MGROUP_ROTATE LRadiator (0, LRadiatorGrp, 2,
		_V(-2.81, 1.3, 0), _V(0,0,1), (float)(-35.5*RAD));

	anim_door = CreateAnimation (0);
	AddAnimationComponent (anim_door, 0.0,    0.3387, &RCargoDoor);
	AddAnimationComponent (anim_door, 0.3656, 0.7043, &LCargoDoor);
	AddAnimationComponent (anim_door, 0.7312, 0.9999, &RRadiator);
	AddAnimationComponent (anim_door, 0.7312, 0.9999, &LRadiator);

	// ***** 2. Landing gear animation *****

	static UINT LNosewheelDoorGrp[2] = {42,50};
	static MGROUP_ROTATE LNosewheelDoor (0, LNosewheelDoorGrp, 2,
		_V(-0.67, -2.15, 17), _V(0, 0.195, 0.981), -1.210f);
	static UINT RNosewheelDoorGrp[2] = {43,51};
	static MGROUP_ROTATE RNosewheelDoor (0, RNosewheelDoorGrp, 2,
		_V(0.78, -2.15, 17), _V(0, 0.195, 0.981), 1.210f);
	static UINT NosewheelGrp[7] = {52,53,54,55,56,57,46};
	static MGROUP_ROTATE Nosewheel (0, NosewheelGrp, 7,
		_V(0.6, -2.0, 18.2), _V(1, 0, 0), 1.8f);
	static UINT RGearDoorGrp[2] = {44,49};
	static MGROUP_ROTATE RGearDoor (0, RGearDoorGrp, 2,
		_V(4.2, -2.64, -1.69), _V(0, 0.02, 0.9), 1.469f);
	static UINT LGearDoorGrp[2] = {45,48};
	static MGROUP_ROTATE LGearDoor (0, LGearDoorGrp, 2,
		_V(-4.2, -2.64, -1.69), _V(0, 0.02, 0.9), -1.469f);
	static UINT MainGearGrp[12] = {47,58,59,60,61,62,63,64,65,66,67,68};
	static MGROUP_ROTATE MainGear (0, MainGearGrp, 12,
		_V(0, -2.66, -3.68), _V(1, 0, 0), 1.65f);

	anim_gear = CreateAnimation (0);
	AddAnimationComponent (anim_gear, 0,   0.5, &LNosewheelDoor);
	AddAnimationComponent (anim_gear, 0,   0.5, &RNosewheelDoor);
	AddAnimationComponent (anim_gear, 0.4, 1.0, &Nosewheel);
	AddAnimationComponent (anim_gear, 0,   0.5, &RGearDoor);
	AddAnimationComponent (anim_gear, 0,   0.5, &LGearDoor);
	AddAnimationComponent (anim_gear, 0.4, 1.0, &MainGear);

	// ***** 3. Ku-band antenna animation *****

	static UINT KuBand1Grp[9] = {72,73,74,75,76,77,78,79,80};
	static MGROUP_ROTATE KuBand1 (0, KuBand1Grp, 9,
		_V(2.85, 0.85, 0), _V(0,0,1), (float)(-18*RAD));
	static UINT KuBand2Grp[6] = {73,74,75,77,78,80};
	static MGROUP_ROTATE KuBand2 (0, KuBand2Grp, 6,
		_V(2.78, 1.7, 0), _V(0,0,1), (float)(-90*RAD));
	static UINT KuBand3Grp[8] = {72,73,74,75,76,77,78,80};
	static MGROUP_ROTATE KuBand3 (0, KuBand3Grp, 8,
		_V(2.75, 2.05, 11.47), _V(0,1,0), (float)(-130*RAD));

	anim_kubd = CreateAnimation (0);
	AddAnimationComponent (anim_kubd, 0,     0.333, &KuBand1);
	AddAnimationComponent (anim_kubd, 0.333, 0.667, &KuBand2);
	AddAnimationComponent (anim_kubd, 0.667, 0.999, &KuBand3);

	// ***** 4. Speedbrake animation *****

	static UINT SB1Grp[1] = {123};
	static MGROUP_ROTATE SB1 (0, SB1Grp, 1,
		_V(0.32,5.77,-12.17), _V(-0.037,0.833,-0.552), (float)(-49.3*RAD));
	static UINT SB2Grp[1] = {124};
	static MGROUP_ROTATE SB2 (0, SB2Grp, 1,
		_V(-0.32,5.77,-12.17), _V(0.037,0.833,-0.552), (float)(49.3*RAD));

	anim_spdb = CreateAnimation (0);
	AddAnimationComponent (anim_spdb, 0, 1, &SB1);
	AddAnimationComponent (anim_spdb, 0, 1, &SB2);

	// ***** 5. RMS arm animation *****
	// Note that the animation components can't be declared static here, since
	// their rotation parameters are modified by the respective parent transforms

	ANIMATIONCOMPONENT_HANDLE parent;

	static UINT RMSShoulderYawGrp[1] = {110};
	rms_anim[0] = new MGROUP_ROTATE (0, RMSShoulderYawGrp, 1,
		_V(-2.26, 1.5, 10), _V(0, 1, 0), (float)(-360*RAD)); // -180 .. +180
	anim_arm_sy = CreateAnimation (0.5);
	parent = AddAnimationComponent (anim_arm_sy, 0, 1, rms_anim[0]);

	static UINT RMSShoulderPitchGrp[1] = {111};
	rms_anim[1] = new MGROUP_ROTATE (0, RMSShoulderPitchGrp, 1,
		_V(-2.26, 1.8, 10), _V(1, 0, 0), (float)(147*RAD)); // -2 .. +145
	anim_arm_sp = CreateAnimation (0.0136);
	parent = AddAnimationComponent (anim_arm_sp, 0, 1, rms_anim[1], parent);

	static UINT RMSElbowPitchGrp[4] = {112,118,119,120};
	rms_anim[2] = new MGROUP_ROTATE (0, RMSElbowPitchGrp, 4,
		_V(-2.26,1.7,3.3), _V(1,0,0), (float)(-162*RAD)); // -160 .. +2
	anim_arm_ep = CreateAnimation (0.0123);
	parent = AddAnimationComponent (anim_arm_ep, 0, 1, rms_anim[2], parent);

	static UINT RMSWristPitchGrp[1] = {113};
	rms_anim[3] = new MGROUP_ROTATE (0, RMSWristPitchGrp, 1,
		_V(-2.26,1.7,-3.55), _V(1,0,0), (float)(240*RAD)); // -120 .. +120
	anim_arm_wp = CreateAnimation (0.5);
	parent = AddAnimationComponent (anim_arm_wp, 0, 1, rms_anim[3], parent);

	static UINT RMSWristYawGrp[2] = {114, 121};
	rms_anim[4] = new MGROUP_ROTATE (0, RMSWristYawGrp, 2,
		_V(-2.26,1.7,-4.9), _V(0,1,0), (float)(-240*RAD)); // -120 .. +120
	anim_arm_wy = CreateAnimation (0.5);
	parent = AddAnimationComponent (anim_arm_wy, 0, 1, rms_anim[4], parent);

	rms_anim[5] = new MGROUP_ROTATE (LOCALVERTEXLIST, MAKEGROUPARRAY(arm_tip), 3,
		_V(-2.26,1.7,-6.5), _V(0,0,1), (float)(894*RAD)); // -447 .. +447
	anim_arm_wr = CreateAnimation (0.5);
	hAC_arm = AddAnimationComponent (anim_arm_wr, 0, 1, rms_anim[5], parent);
}

void Atlantis::AddOrbiterVisual (const VECTOR3 &ofs)
{
	orbiter_ofs = ofs;

	UINT meshidx;
	meshidx = AddMesh (hOrbiterMesh, &ofs);
	SetMeshVisibilityMode (meshidx, render_cockpit ? MESHVIS_ALWAYS : MESHVIS_EXTERNAL);

	if (do_cargostatic) {
		VECTOR3 cofs = ofs + cargo_static_ofs;
		AddMesh (cargo_static_mesh_name, &cofs);
	}
	if (do_plat) {
		VECTOR3 plat_ofs = _V(-2.59805, 1.69209, -5.15524);
		AddMesh("shuttle_eva_plat", &plat_ofs);
	}

	SetCameraOffset (_V(ofs.x-0.55,ofs.y+2.4,ofs.z+15.0));

	if (!sat_attach) sat_attach = CreateAttachment (false, ofs+ofs_sts_sat, _V(0,1,0), _V(0,0,1), "X");
	else SetAttachmentParams (sat_attach, ofs+ofs_sts_sat, _V(0,1,0), _V(0,0,1));

	if (!rms_attach) rms_attach = CreateAttachment (false, ofs+arm_tip[0], arm_tip[1]-arm_tip[0], arm_tip[2]-arm_tip[0], "G", true);
	else SetAttachmentParams (rms_attach, ofs+arm_tip[0], arm_tip[1]-arm_tip[0], arm_tip[2]-arm_tip[0]);
}

void Atlantis::AddTankVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hTankMesh, &ofs);
	SetMeshVisibilityMode (meshidx, MESHVIS_ALWAYS);
}

void Atlantis::AddSRBVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hSRBMesh, &ofs);
	SetMeshVisibilityMode (meshidx, MESHVIS_ALWAYS);
	int id = AddExhaustRef (EXHAUST_CUSTOM, _V(ofs.x, ofs.y, ofs.z-21.8), 0, 0);
	if (ofs.x < 0.0) srb_id1 = id;
	else             srb_id2 = id;
}

void Atlantis::SeparateBoosters (double met)
{
	// Create SRB's as individual objects
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.flag[0] = 3;
	GetSRB_State (met, vs.eng_main, vs.fuel);
	Local2Rel (OFS_LAUNCH_RIGHTSRB, vs.rpos);
	vs.arot.z += 0.25*PI;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-SRB1");
	oapiCreateVessel (name, "Atlantis_SRB", vs);
	Local2Rel (OFS_LAUNCH_LEFTSRB, vs.rpos);
	vs.arot.z += 0.5*PI;
	name[strlen(name)-1] = '2';
	oapiCreateVessel (name, "Atlantis_SRB", vs);

	// Remove SRB's from Shuttle instance
	DelPropellantResource (ph_srb);
	DelThrusterGroup (thg_srb, THGROUP_USER, true);

	// reconfigure
	SetOrbiterTankConfiguration();
	ShiftCentreOfMass (OFS_LAUNCH_ORBITER-OFS_WITHTANK_ORBITER);
}

void Atlantis::SeparateTank (void)
{
	// Create Tank as individual object
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.eng_main = vs.eng_hovr = 0.0;
	VECTOR3 ofs = OFS_WITHTANK_TANK;
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	VECTOR3 vel = {0,-1,0};
	Local2Rel (ofs, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.x = -0.02;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-Tank");
	oapiCreateVessel (name, "Atlantis_Tank", vs);

	// Remove Tank from shuttle instance
	DelPropellantResource (ph_tank);

	// main engines are done
	DelThrusterGroup (thg_main, THGROUP_MAIN, true);

	// clear launch attitude control system
	DelThrusterGroup (THGROUP_ATT_PITCHUP, true);
	DelThrusterGroup (THGROUP_ATT_PITCHDOWN, true);
	DelThrusterGroup (THGROUP_ATT_BANKLEFT, true);
	DelThrusterGroup (THGROUP_ATT_BANKRIGHT, true);
	DelThrusterGroup (THGROUP_ATT_YAWLEFT, true);
	DelThrusterGroup (THGROUP_ATT_YAWRIGHT, true);

	// reconfigure
	SetOrbiterConfiguration ();
	ShiftCentreOfMass (OFS_WITHTANK_ORBITER);
}

void Atlantis::ToggleGrapple (void)
{
	HWND hDlg;
	OBJHANDLE hV = GetAttachmentStatus (rms_attach);

	if (hV) {  // release satellite

		ATTACHMENTHANDLE hAtt = CanArrest();
		DetachChild (rms_attach);
		if (hDlg = oapiFindDialog (g_Param.hDLL, IDD_RMS)) {
			SetWindowText (GetDlgItem (hDlg, IDC_GRAPPLE), "Grapple");
			EnableWindow (GetDlgItem (hDlg, IDC_STOW), TRUE);
		}
		// check whether the object being ungrappled is ready to be clamped into the payload bay
		if (hAtt) {
			AttachChild (hV, sat_attach, hAtt);
			if (hDlg) {
				SetWindowText (GetDlgItem (hDlg, IDC_PAYLOAD), "Purge");
				EnableWindow (GetDlgItem (hDlg, IDC_PAYLOAD), TRUE);
			}
		}

#ifdef UNDEF
		VECTOR3 pos, dir, rot, gbay, gpos;
		GetAttachmentParams (sat_attach, pos, dir, rot);
		Local2Global (pos, gbay);
		VESSEL *v = oapiGetVesselInterface (hV);
		DWORD nAttach = v->AttachmentCount (true);
		for (DWORD j = 0; j < nAttach; j++) { // now scan all attachment points
			ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle (true, j);
			v->GetAttachmentParams (hAtt, pos, dir, rot);
			v->Local2Global (pos, gpos);
			if (dist (gpos, gbay) < MAX_GRAPPLING_DIST) {
				AttachChild (hV, sat_attach, hAtt);
				return;
			}
		}
#endif

	} else {             // grapple satellite

		VECTOR3 gpos, grms, pos, dir, rot;
		Local2Global (orbiter_ofs+arm_tip[0], grms);  // global position of RMS tip
		
		// Search the complete vessel list for a grappling candidate.
		// Not very scalable ...
		for (DWORD i = 0; i < oapiGetVesselCount(); i++) {
			OBJHANDLE hV = oapiGetVesselByIndex (i);
			if (hV == GetHandle()) continue; // we don't want to grapple ourselves ...
			oapiGetGlobalPos (hV, &gpos);
			if (dist (gpos, grms) < oapiGetSize (hV)) { // in range
				VESSEL *v = oapiGetVesselInterface (hV);
				DWORD nAttach = v->AttachmentCount (true);
				for (DWORD j = 0; j < nAttach; j++) { // now scan all attachment points of the candidate
					ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle (true, j);
					const char *id = v->GetAttachmentId (hAtt);
					if (strncmp (id, "GS", 2)) continue; // attachment point not compatible
					v->GetAttachmentParams (hAtt, pos, dir, rot);
					v->Local2Global (pos, gpos);
					if (dist (gpos, grms) < MAX_GRAPPLING_DIST) { // found one!
						// check whether satellite is currently clamped into payload bay
						if (hV == GetAttachmentStatus (sat_attach))
							DetachChild (sat_attach);
						AttachChild (hV, rms_attach, hAtt);
						if (hDlg = oapiFindDialog (g_Param.hDLL, IDD_RMS)) {
							SetWindowText (GetDlgItem (hDlg, IDC_GRAPPLE), "Release");
							EnableWindow (GetDlgItem (hDlg, IDC_STOW), FALSE);
						}
						return;
					}
				}
			}
		}

	}
}

void Atlantis::ToggleArrest (void)
{
	HWND hDlg;
	if (SatStowed()) { // purge satellite
		DetachChild (sat_attach, 0.1);
		if (hDlg = oapiFindDialog (g_Param.hDLL, IDD_RMS)) {
			SetWindowText (GetDlgItem (hDlg, IDC_PAYLOAD), "Arrest");
			EnableWindow (GetDlgItem (hDlg, IDC_PAYLOAD), CanArrest() ? TRUE:FALSE);
		}
	} else if (CanArrest()) {           // try to arrest satellite
		ToggleGrapple();
	}
}

// check whether the currently grappled object can be stowed in the cargo bay
ATTACHMENTHANDLE Atlantis::CanArrest (void) const
{
	OBJHANDLE hV = GetAttachmentStatus (rms_attach);
	if (!hV) return 0;
	VESSEL *v = oapiGetVesselInterface (hV);
	DWORD nAttach = v->AttachmentCount (true);
	VECTOR3 pos, dir, rot, gpos, gbay;
	GetAttachmentParams (sat_attach, pos, dir, rot);
	Local2Global (pos, gbay);
	for (DWORD j = 0; j < nAttach; j++) {
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle (true, j);
		if (strncmp (v->GetAttachmentId (hAtt), "XS", 2)) continue; // attachment point not compatible
		v->GetAttachmentParams (hAtt, pos, dir, rot);
		v->Local2Global (pos, gpos);
		if (dist (gpos, gbay) < MAX_GRAPPLING_DIST) {
			return hAtt;
		}
	}
	return 0;
}

void Atlantis::SeparateMMU (void)
{
	// Create Tank as individual object
	VESSELSTATUS vs;
	GetStatus (vs);
	char name[256];
	strcpy (name, GetName()); strcat (name, "-MMU");
	hMMU = VESSEL::Create (name, "Nasa_MMU", vs);
	jettison_time = oapiGetSimTime();
	reset_mmu = true;
	// Remove Tank from shuttle instance
	SetOrbiterConfiguration ();
}

void Atlantis::AutoMainGimbal ()
{
	// Implement automatic gimbal adjustment for main engines to correct for
	// SRB thrust variations. We adjust only the upper engine

	double        F_srb  = SRB_THRUST         *GetThrusterLevel (th_srb[0]);
	double        F_main = ORBITER_MAIN_THRUST*GetThrusterLevel (th_main[0]);
	if (F_main) {
		double        M_srb =   2.51951112*F_srb;  // angular moment from SRB thrust
		double        M_m2  = -12.02495   *F_main; // angular moment from the two lower main engines in neutral position
		double        M_0   = M_srb + M_m2;
		static double ry    =   9.42;  // upper main thruster: y-offset
		static double rz    = -23.295; // upper main thruster: z-offset
		static double ry2   = ry*ry;
		static double rz2   = rz*rz;
		double        F2    = F_main*F_main;
		double        M2    = M_0*M_0;
		double        term  = -M2 + F2*(ry2+rz2);
		double        arg1  = max (0.0, rz2 * term);
		double        scale = 1.0/(F_main*(ry2+rz2));
		double        dz    = (M_0*ry + sqrt (arg1))*scale;
		double		  arg2  = max (0.0, ry2 * term);
		double        dy    =-(M_0*rz + sqrt (arg2))*scale;
		if (!arg1 || !arg2) {
			double len = hypot(dy,dz);
			dy /= len, dz /= len;
		}
		SetThrusterDir (th_main[2], _V(0,dy,dz));
	}
}

void Atlantis::LaunchClamps ()
{
	VECTOR3 F, T, r = _V(0,0,0), Fc = _V(0,0,0), Tc = _V(0,0,0);
	GetThrusterMoment (th_srb[0], F, T);
	Fc.z = -2*F.z;
	Tc.x =  2*T.x;
	GetThrusterMoment (th_main[0], F, T);
	Fc.z -= 2*F.z;
	Fc.y -= 2*F.y;
	Tc.x += 2*T.x;
	GetThrusterMoment (th_main[2], F, T);
	Fc.z -= F.z;
	Fc.y -= F.y;
	Tc.x += T.x;
	r.z = (Fc.y ? Tc.x/Fc.y : 0);
	AddForce (Fc, r);
}

void Atlantis::SetGearParameters (double state)
{
	if (state == 1.0) { // gear fully deployed
		SetTouchdownPoints (_V(0,-4,17), _V(-3.96,-5.6,-4.3), _V(3.96,-5.6,-4.3)); // gear wheel tips
		SetSurfaceFrictionCoeff (0.05, 0.4);
	} else {
		SetTouchdownPoints (_V(0,-2.5,14), _V(-8,-2.8,-9), _V(8,-2.8,-9)); // belly landing
		SetSurfaceFrictionCoeff (0.4, 0.4);
	}
}

void Atlantis::Timestep (double simt)
{
	double met;
	int i;
	OBJHANDLE hvessel;

	switch (status) {
	case 0: // launch configuration
		if (GetEngineLevel (ENGINE_MAIN) > 0.95) {
			status = 1; // launch
			t0 = simt + SRB_STABILISATION_TIME;   // store designated liftoff time
		} else
			AutoMainGimbal();
		break;
	case 1: // SRB's ignited
		met = simt-t0;
		if (met > SRB_SEPARATION_TIME || bManualSeparate) { // separate boosters
			SeparateBoosters (met);
			bManualSeparate = false;
		} else {
			// extract current thrust level and propellant level as a function of time
			double thrust_level, prop_level;
			GetSRB_State (met, thrust_level, prop_level);
			for (i = 0; i < 2; i++)
				SetThrusterLevel (th_srb[i], thrust_level);
			if (met < 0.0) LaunchClamps ();
			AutoMainGimbal();
		}
		break;

	case 2: // post SRB separation
		if (GetPropellantMass (ph_tank) < 1.0 || bManualSeparate) {
			SeparateTank();
			bManualSeparate = false;
		}
		break;
	case 3: // post tank separation
		if (bManualSeparate && GetAttachmentStatus (sat_attach)) {
			DetachChild (sat_attach, 0.1);
			bManualSeparate = false;
		}

		if (do_eva) {
			char name[256];
			strcpy (name, GetName()); strcat (name, "-MMU");
			hvessel=oapiGetVesselByName(name);
			if (hvessel == 0)
			{
				SeparateMMU();
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			}
			else
			{
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			};
			do_eva = false;
		};
		break;
	}

	VESSEL *aVessel;
	VESSELSTATUS vs;

	if (reset_mmu && simt-jettison_time > .01)
	{
		GetStatus(vs);
		vs.eng_main = vs.eng_hovr = 0.0;
		VECTOR3 ofs = OFS_ZERO;
		ofs.x += OFS_MMU.x;
		ofs.y += OFS_MMU.y;
		ofs.z += OFS_MMU.z;
		VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
		VECTOR3 vel = {0,0,0};
		Local2Rel (ofs, vs.rpos);
		GlobalRot (vel, rofs);
		vs.rvel.x = rvel.x+rofs.x;
		vs.rvel.y = rvel.y+rofs.y;
		vs.rvel.z = rvel.z+rofs.z;
		aVessel = new VESSEL (hMMU, 1);
		aVessel->DefSetState(&vs);
		reset_mmu=false;
	}

	// ***** Animate cargo doors *****

	if (door_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * DOOR_OPERATING_SPEED;
		if (door_status == DOOR_CLOSING) {
			if (door_proc > 0.0) door_proc = max (0.0, door_proc-da);
			else                 door_status = DOOR_CLOSED;
		} else { // door opening
			if (door_proc < 1.0) door_proc = min (1.0, door_proc+da);
			else                 door_status = DOOR_OPEN;
		}
		SetAnimation (anim_door, door_proc);
		rdoor_drag = sqrt (min (1.0, door_proc*3.0));
		ldoor_drag = sqrt (min (1.0, max(0.0, door_proc-0.3656)*3.0));
	}

	// ***** Animate landing gear *****

	if (gear_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
		if (gear_status == DOOR_CLOSING) { // retract gear
			if (gear_proc > 0.0) gear_proc = max (0.0, gear_proc-da);
			else                 gear_status = DOOR_CLOSED;
		} else {                           // deploy gear
			if (gear_proc < 1.0) gear_proc = min (1.0, gear_proc+da);
			else                 gear_status = DOOR_OPEN;
		}
		SetAnimation (anim_gear, gear_proc);
		SetGearParameters (gear_proc);
	}

	// ***** Animate Ku-band antenna *****

	if (kubd_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * KU_OPERATING_SPEED;
		if (kubd_status == DOOR_CLOSING) { // retract antenna
			if (kubd_proc > 0.0) kubd_proc = max (0.0, kubd_proc-da);
			else                 kubd_status = DOOR_CLOSED;
		} else {                           // deploy antenna
			if (kubd_proc < 1.0) kubd_proc = min (1.0, kubd_proc+da);
			else                 kubd_status = DOOR_OPEN;
		}
		SetAnimation (anim_kubd, kubd_proc);
	}

	// ***** Animate speedbrake *****

	if (spdb_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * SPEEDBRAKE_OPERATING_SPEED;
		if (spdb_status == DOOR_CLOSING) { // retract brake
			if (spdb_proc > 0.0) spdb_proc = max (0.0, spdb_proc-da);
			else                 spdb_status = DOOR_CLOSED;
		} else {                           // deploy antenna
			if (spdb_proc < 1.0) spdb_proc = min (1.0, spdb_proc+da);
			else                 spdb_status = DOOR_OPEN;
		}
		SetAnimation (anim_spdb, spdb_proc);
	}

	// ***** Stow RMS arm *****

	if (center_arm) {
		double t0 = oapiGetSimTime();
		double dt = t0 - center_arm_t;       // time step
		double da = ARM_OPERATING_SPEED*dt;  // total rotation angle

		// work from the wrist down to the shoulder
		if (da && (arm_wr != 0.5)) {    // zero wrist roll
			if (da >= fabs(arm_wr-0.5)) // finished
				arm_wr = 0.5, da -= fabs(arm_wr-0.5);
			else
				arm_wr -= (arm_wr > 0.5 ? da:-da), da = 0;
			SetAnimationArm (anim_arm_wr, arm_wr);
		}
		if (da && (arm_wy != 0.5)) {    // zero wrist yaw
			if (da >= fabs(arm_wy-0.5)) // finished
				arm_wy = 0.5, da -= fabs(arm_wy-0.5);
			else
				arm_wy -= (arm_wy > 0.5 ? da:-da), da = 0;
			SetAnimationArm (anim_arm_wy, arm_wy);
		}
		if (da && (arm_wp != 0.5)) {    // zero wrist pitch
			if (da >= fabs(arm_wp-0.5)) // finished
				arm_wp = 0.5, da -= fabs(arm_wp-0.5);
			else
				arm_wp -= (arm_wp > 0.5 ? da:-da), da = 0;
			SetAnimationArm (anim_arm_wp, arm_wp);
		}
		if (da && arm_ep) {             // zero elbow pitch
			if (da >= arm_ep)           // finished
				arm_ep = 0.0, da -= arm_ep;
			else
				arm_ep -= da, da = 0;
			SetAnimationArm (anim_arm_ep, arm_ep);
		}
		if (da && (arm_sy != 0.5)) {    // zero shoulder yaw
			if (da >= fabs(arm_sy-0.5)) // finished
				arm_sy = 0.5, da -= fabs(arm_sy-0.5);
			else
				arm_sy -= (arm_sy > 0.5 ? da:-da), da = 0;
			SetAnimationArm (anim_arm_sy, arm_sy);
		}
		if (da && arm_sp) {             // zero shoulder pitch
			if (da >= arm_sp)           // finished
				arm_sp = 0.0, da -= arm_sp;
			else
				arm_sp -= da, da = 0;
			SetAnimationArm (anim_arm_sp, arm_sp);
		}
		center_arm_t = t0;
		if (da) center_arm = false; // finished stowing
	}

	if (arm_moved) {
		SetAttachmentParams (rms_attach, orbiter_ofs+arm_tip[0], arm_tip[1]-arm_tip[0], arm_tip[2]-arm_tip[0]);
		arm_moved = false;
	}

}

void Atlantis::Jettison ()
{
	switch (status) {
	case 0:
	case 3:               // nothing to do
		break;
	case 1:               // abandon boosters
		SeparateBoosters (oapiGetSimTime()-t0); 
		break;
	case 2:               // abandon tank
		SeparateTank();
		break;
	}
}

// Update moving parts of the orbiter's visual: payload bay doors and gear
// This should only be called when the visual exists, e.g. from within
// ovcVisualCreated or ovcAnimate

void Atlantis::UpdateMesh ()
{
	// update animation states
	SetAnimation (anim_door, door_proc);
	SetAnimation (anim_gear, gear_proc);
	SetAnimation (anim_kubd, kubd_proc);
	SetAnimation (anim_spdb, spdb_proc);

	SetAnimationArm (anim_arm_sy, arm_sy);
	SetAnimationArm (anim_arm_sp, arm_sp);
	SetAnimationArm (anim_arm_ep, arm_ep);
	SetAnimationArm (anim_arm_wp, arm_wp);
	SetAnimationArm (anim_arm_wy, arm_wy);
	SetAnimationArm (anim_arm_wr, arm_wr);
}

void Atlantis::OperateLandingGear (DoorStatus action)
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	gear_status = action;
}

void Atlantis::RevertLandingGear ()
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	OperateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void Atlantis::OperateCargoDoors (DoorStatus action)
{
	if (kubd_status != DOOR_CLOSED) return;
	// operate payload bay doors only if Ku-band antenna is stowed

	door_status = action;
}

void Atlantis::RevertCargoDoors (void)
{
	OperateCargoDoors (door_status == DOOR_CLOSED || door_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void Atlantis::OperateKuBand (DoorStatus action)
{
	if (door_status != DOOR_OPEN) return;
	// operate Ku band antenna only if payload bay doors are open

	kubd_status = action;
}

void Atlantis::RevertKuBand (void)
{
	OperateKuBand (kubd_status == DOOR_CLOSED || kubd_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void Atlantis::OperateSpeedbrake (DoorStatus action)
{
	spdb_status = action;
}

void Atlantis::RevertSpeedbrake (void)
{
	OperateSpeedbrake (spdb_status == DOOR_CLOSED || spdb_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void Atlantis::SetAnimationArm (UINT anim, double state)
{
	SetAnimation (anim, state);
	arm_moved = true;

	HWND hDlg;
	if (!SatStowed() && (hDlg = oapiFindDialog (g_Param.hDLL, IDD_RMS)))
		EnableWindow (GetDlgItem (hDlg, IDC_PAYLOAD), CanArrest() ? TRUE : FALSE);
}

// ==============================================================
// DLL entry point
// ==============================================================

BOOL WINAPI DllMain (HINSTANCE hModule,
					 DWORD ul_reason_for_call,
					 LPVOID lpReserved)
{
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		g_Param.hDLL = hModule;
		break;
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}

// ==============================================================
// API interface
// ==============================================================

// Vessel initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new Atlantis (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (Atlantis*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	((Atlantis*)vessel)->SetClassCaps (cfg);
}

// Read status from scenario file
DLLCLBK void ovcLoadStateEx (VESSEL *vessel, FILEHANDLE scn, void *vs)
{
	((Atlantis*)vessel)->LoadState (scn, vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((Atlantis*)vessel)->SaveState (scn);
}

DLLCLBK void ovcFocusChanged (VESSEL *vessel, bool getfocus, OBJHANDLE newv, OBJHANDLE oldv)
{
	if (getfocus) {
		oapiDisableMFDMode (MFD_LANDING);
	}
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->Timestep (simt);
}

DLLCLBK void ovcAnimate (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->UpdateMesh ();
}

DLLCLBK void ovcVisualCreated (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	if (refcount > 1) return; // we don't allow more than one visual per object
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = vis;

	// reset grappling point
	sts->arm_tip[0] = _V(-2.26,1.71,-6.5);
	sts->arm_tip[1] = _V(-2.26,1.71,-7.5);
	sts->arm_tip[2] = _V(-2.26,2.71,-6.5);

	sts->UpdateMesh ();
}

DLLCLBK void ovcVisualDestroyed (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = NULL;
}

// Keyboard interface handler (buffered key events)
DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	Atlantis *sts = (Atlantis*)vessel;

	if (KEYMOD_SHIFT (kstate)) {

		switch (key) {
		case OAPI_KEY_J: // "Jettison"
			sts->bManualSeparate = true;
			return 1;
#ifdef UNDEF
		case OAPI_KEY_1: 
			if (sts->arm_shoulder_pitch - sts->arm_angle > ARM_SHOULDER_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;		
		case OAPI_KEY_2: 
			if (sts->arm_shoulder_yaw - sts->arm_angle > ARM_SHOULDER_YAW_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_YAW;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_3:
			if (sts->arm_elbow_pitch - sts->arm_angle > ARM_ELBOW_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_ELBOW_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_4:
			if (sts->arm_wrist_pitch - sts->arm_angle > ARM_WRIST_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_5:
			if (sts->arm_wrist_yaw - sts->arm_angle > ARM_WRIST_YAW_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_YAW;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_6: 
			if (sts->arm_wrist_roll - sts->arm_angle > ARM_WRIST_ROLL_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_ROLL;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
#endif
		case OAPI_KEY_E:
			if (sts->status != 3) return 1; // Allow MMU only after orbiter has detached from MT
#ifdef UNDEF
			double total_rotation=0;
			total_rotation += sts->arm_shoulder_pitch;
			total_rotation += sts->arm_shoulder_yaw;
			total_rotation += sts->arm_elbow_pitch;
			total_rotation += sts->arm_wrist_pitch;
			total_rotation += sts->arm_wrist_yaw;
			total_rotation += sts->arm_wrist_roll;
			if (total_rotation ==0)
			{
				if (sts->do_plat)
				{
					sts->do_plat=false;
				}
				else
				{
					sts->do_plat=true;
				}
				sts->ClearMeshes();
				sts->ClearExhaustRefs();
				sts->ClearAttExhaustRefs();
				sts->AddOrbiterVisual (OFS_ZERO);
			}
#endif
			return 1;
		}	
	} else if (KEYMOD_CONTROL (kstate)) {
		switch (key) {
		case OAPI_KEY_SPACE: // open RMS control dialog
			oapiOpenDialog (g_Param.hDLL, IDD_RMS, RMS_DlgProc, sts);
			return 1;
		case OAPI_KEY_B: // deploy/retract speedbrake
			sts->RevertSpeedbrake ();
			return 1;
		case OAPI_KEY_U: // deploy/store Ku-band antenna
			sts->RevertKuBand ();
			return 1;
		}
	} else { // unmodified keys
		switch (key) {
		case OAPI_KEY_G:  // "Landing gear"
			sts->RevertLandingGear ();
			return 1;
		case OAPI_KEY_J:  // "Jettison"
			sts->bManualSeparate = true;
			return 1;
		case OAPI_KEY_K:  // "Cargo bay doors"
			sts->RevertCargoDoors ();
			return 1;
#ifdef UNDEF
		case OAPI_KEY_1:
			if (sts->arm_shoulder_pitch + sts->arm_angle < ARM_SHOULDER_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_2: 
			if (sts->arm_shoulder_yaw + sts->arm_angle < ARM_SHOULDER_YAW_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_YAW;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_3: 
			if (sts->arm_elbow_pitch + sts->arm_angle < ARM_ELBOW_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_ELBOW_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_4: 
			if (sts->arm_wrist_pitch + sts->arm_angle < ARM_WRIST_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_5: 
			if (sts->arm_wrist_yaw + sts->arm_angle < ARM_WRIST_YAW_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_YAW;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_6:
			if (sts->arm_wrist_roll + sts->arm_angle < ARM_WRIST_ROLL_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_ROLL;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
#endif
		case OAPI_KEY_8:
			sts->ToggleGrapple();
			return 1;
		case OAPI_KEY_9: 
			sts->center_arm = true;
			return 1;
#ifdef UNDEF
		case OAPI_KEY_0: 
			if (sts->arm_angle == ARM_ANGLE_HIGH)
				sts->arm_angle = ARM_ANGLE_MED;
			else if (sts->arm_angle == ARM_ANGLE_MED)
				sts->arm_angle = ARM_ANGLE_LOW;
			else if (sts->arm_angle == ARM_ANGLE_LOW)
				sts->arm_angle = ARM_ANGLE_HIGH;
			return 1;
#endif
		case OAPI_KEY_E: 
			sts->do_eva=true;
			return 1;
		}
	}
	return 0;
}

// Message callback function for RMS control dialog box

BOOL CALLBACK RMS_DlgProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	Atlantis *sts = (Atlantis*)oapiGetDialogContext (hWnd);
	// pointer to vessel instance was passed as dialog context

	const double step = 0.05*RAD;
	static double t0;
	double t1;
	HICON hIcon;

	switch (uMsg) {
	case WM_INITDIALOG:
		sts = (Atlantis*)lParam;
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_UP));
		SendDlgItemMessage (hWnd, IDC_WRIST_PITCHUP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_ELBOW_PITCHUP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_SHOULDER_PITCHUP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_DOWN));
		SendDlgItemMessage (hWnd, IDC_WRIST_PITCHDOWN, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_ELBOW_PITCHDOWN, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_SHOULDER_PITCHDOWN, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_LEFT));
		SendDlgItemMessage (hWnd, IDC_WRIST_YAWLEFT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_SHOULDER_YAWLEFT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_RIGHT));
		SendDlgItemMessage (hWnd, IDC_WRIST_YAWRIGHT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_SHOULDER_YAWRIGHT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_RRIGHT));
		SendDlgItemMessage (hWnd, IDC_WRIST_ROLLRIGHT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		hIcon = LoadIcon (g_Param.hDLL, MAKEINTRESOURCE(IDI_RLEFT));
		SendDlgItemMessage (hWnd, IDC_WRIST_ROLLLEFT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);
		SendDlgItemMessage (hWnd, IDC_SHOWGRAPPLE, BM_SETCHECK, oapiGetShowGrapplePoints() ? BST_CHECKED:BST_UNCHECKED, 0);
		SetWindowText (GetDlgItem (hWnd, IDC_GRAPPLE), sts->SatGrappled() ? "Release" : "Grapple");
		EnableWindow (GetDlgItem (hWnd, IDC_STOW), sts->SatGrappled() ? FALSE : TRUE);
		SetWindowText (GetDlgItem (hWnd, IDC_PAYLOAD), sts->SatStowed() ? "Purge" : "Arrest");
		EnableWindow (GetDlgItem (hWnd, IDC_PAYLOAD), sts->SatStowed() || sts->CanArrest() ? TRUE:FALSE);
		SetTimer (hWnd, 1, 50, NULL);
		t0 = oapiGetSimTime();
		return FALSE;
	case WM_DESTROY:
		KillTimer (hWnd, 1);
		return 0;
	case WM_TIMER:
		t1 = oapiGetSimTime();
		if (SendDlgItemMessage (hWnd, IDC_SHOULDER_YAWLEFT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_sy = min (1.0, sts->arm_sy + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_sy, sts->arm_sy);
		} else if (SendDlgItemMessage (hWnd, IDC_SHOULDER_YAWRIGHT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_sy = max (0.0, sts->arm_sy - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_sy, sts->arm_sy);
		} else if (SendDlgItemMessage (hWnd, IDC_SHOULDER_PITCHUP, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_sp = min (1.0, sts->arm_sp + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_sp, sts->arm_sp);
		} else if (SendDlgItemMessage (hWnd, IDC_SHOULDER_PITCHDOWN, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_sp = max (0.0, sts->arm_sp - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_sp, sts->arm_sp);
		} else if (SendDlgItemMessage (hWnd, IDC_ELBOW_PITCHUP, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_ep = max (0.0, sts->arm_ep - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_ep, sts->arm_ep);
		} else if (SendDlgItemMessage (hWnd, IDC_ELBOW_PITCHDOWN, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_ep = min (1.0, sts->arm_ep + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_ep, sts->arm_ep);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_PITCHUP, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wp = min (1.0, sts->arm_wp + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wp, sts->arm_wp);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_PITCHDOWN, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wp = max (0.0, sts->arm_wp - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wp, sts->arm_wp);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_YAWLEFT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wy = min (1.0, sts->arm_wy + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wy, sts->arm_wy);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_YAWRIGHT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wy = max (0.0, sts->arm_wy - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wy, sts->arm_wy);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_ROLLLEFT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wr = max (0.0, sts->arm_wr - (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wr, sts->arm_wr);
		} else if (SendDlgItemMessage (hWnd, IDC_WRIST_ROLLRIGHT, BM_GETSTATE, 0, 0) & BST_PUSHED) {
			sts->arm_wr = min (1.0, sts->arm_wr + (t1-t0)*ARM_OPERATING_SPEED);
			sts->SetAnimationArm (sts->anim_arm_wr, sts->arm_wr);
		}
		t0 = t1;
		return 0;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDCANCEL:
			oapiCloseDialog (hWnd);
			return TRUE;
		case IDC_STOW:
			if (sts->center_arm = !sts->center_arm)
				sts->center_arm_t = oapiGetSimTime();
			return 0;
		case IDC_GRAPPLE:
			sts->ToggleGrapple();
			return 0;
		case IDC_PAYLOAD:
			sts->ToggleArrest();
			return 0;
		case IDC_SHOWGRAPPLE:
			oapiSetShowGrapplePoints (SendDlgItemMessage (hWnd, IDC_SHOWGRAPPLE, BM_GETCHECK, 0, 0) == BST_CHECKED ? true : false);
			return 0;
		}
		break;
	}
	return oapiDefDialogProc (hWnd, uMsg, wParam, lParam);
}