// ==============================================================
//                ORBITER MODULE: DeltaGlider
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2003 Martin Schweiger
//                   All rights reserved
//
// DeltaGlider.cpp
// Reference implementation of delta glider vessel class module
// ==============================================================

#define STRICT

#include "DeltaGlider.h"
#include <stdio.h>
#include <math.h>

#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))

// ==============================================================
// Global parameters
// ==============================================================

GDIParams g_Param;

// ==============================================================
// Airfoil coefficient functions
// Return lift, moment and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// ==============================================================

// 1. vertical lift component (wings and body)

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,    -0.4,      0,    0.7,     1,   0.8,     0,      0};
	static const double CM[nlift]  = {       0,      0,   0.014, 0.0034,-0.0074,-0.094,-0.012,     0,      0};
	for (int i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;
	// aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;
	// aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.018 + 0.4*saoa*saoa; // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (int i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.02 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

// ==============================================================
// Specialised vessel class DeltaGlider
// ==============================================================

// Constructor
DeltaGlider::DeltaGlider (OBJHANDLE hObj, int fmodel)
: VESSEL (hObj, fmodel)
{
	int i;

	aoa_ind = PI;
	slip_ind = PI*0.5;
	slope = slope_ind = PI;
	yg_mode           = 0;
	yg_comp           = 0;
	pg_mode           = 0;
	gear_status       = DOOR_CLOSED;
	gear_proc         = 0.0;
	nose_status       = DOOR_CLOSED;
	nose_proc         = 0.0;
	lock_status       = DOOR_CLOSED;
	lock_proc         = 0.0;
	brake_status      = DOOR_CLOSED;
	brake_proc        = 0.0;
	radiator_status   = DOOR_CLOSED;
	radiator_proc     = 0.0;
	DefineAnimations();
	for (i = 0; i < nsurf; i++) srf[i] = 0;
}

// Destructor
DeltaGlider::~DeltaGlider ()
{
	ReleaseSurfaces();
}

// Set vessel class parameters
void DeltaGlider::SetClassCaps (FILEHANDLE cfg)
{
	// *********************** physical parameters *********************************

	SetSize (9.0);
	SetEmptyMass (12000.0);
	SetCW (0.09, 0.09, 2, 1.4);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(53.0,186.9,25.9));
	SetSurfaceFrictionCoeff (0.07, 0.3);
	SetMaxWheelbrakeForce (5e5);
	SetPMI (_V(15.5,22.1,7.7));
	SetTrimScale (0.05);
	SetCameraOffset (_V(0,1.4,6.8));
	SetDockParams (_V(0,-0.49,9.84), _V(0,0,1), _V(0,1,0));
	SetTouchdownPoints (_V(0,-2.57,10), _V(-3.5,-2.57,-3), _V(3.5,-2.57,-3));
	EnableTransponder (true);

	// **************************** NAV radios *************************************

	InitNavRadios (4);

	// ************************* propellant specs **********************************

	ph_main = CreatePropellantResource (MAX_MAIN_FUEL); // main tank
	ph_rcs  = CreatePropellantResource (MAX_RCS_FUEL);  // RCS tank

	// *********************** thruster definitions ********************************

	PARTICLESTREAMSPEC contrail = {
		0, 8.0, 3, 150, 0.3, 3.5, 4, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	PARTICLESTREAMSPEC exhaust_main = {
		0, 2.0, 10, 150, 0.1, 0.2, 16, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	PARTICLESTREAMSPEC exhaust_hover = {
		0, 1.5, 20, 150, 0.1, 0.15, 9, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};

	// main thrusters
	th_main[0] = CreateThruster (_V(-1,0.0,-7.7), _V(0,0,1), MAX_MAIN_THRUST, ph_main, ISP);
	th_main[1] = CreateThruster (_V( 1,0.0,-7.7), _V(0,0,1), MAX_MAIN_THRUST, ph_main, ISP);
	thg_main = CreateThrusterGroup (th_main, 2, THGROUP_MAIN);
	AddExhaust (th_main[0], 12, 1);
	AddExhaust (th_main[1], 12, 1);
	AddExhaustStream (th_main[0], _V(-1,0,-15), &contrail);
	AddExhaustStream (th_main[1], _V( 1,0,-15), &contrail);
	AddExhaustStream (th_main[0], _V(-1,0,-10), &exhaust_main);
	AddExhaustStream (th_main[1], _V( 1,0,-10), &exhaust_main);

	// retro thrusters
	// note that we have to tilt retros slightly downwards to avoid inducing
	// an angular momentum, since they are mounted below the level of CG.
	// This also means that retros will induce an upward linear component.
	th_retro[0] = CreateThruster (_V(-3,-0.3,5.6), _V(0,0.0535714,-1), MAX_RETRO_THRUST, ph_main, ISP);
	th_retro[1] = CreateThruster (_V( 3,-0.3,5.6), _V(0,0.0535714,-1), MAX_RETRO_THRUST, ph_main, ISP);
	thg_retro = CreateThrusterGroup (th_retro, 2, THGROUP_RETRO);
	AddExhaust (th_retro[0], 3, 0.4);
	AddExhaust (th_retro[1], 3, 0.4);

	// hover thrusters (simplified)
	// This is an example of how multiple engines can be combined into a single
	// "logical" engine if they are always engaged in a fixed configuration.
	// The exhaust render definitions then need to explicitly define their reference
	// positions and directions, since they can't be obtained from the thruster specs
	th_hover[0] = CreateThruster (_V(0,0,0), _V(0,1,0), MAX_HOVER_THRUST, ph_main, ISP);
	thg_hover = CreateThrusterGroup (th_hover, 1, THGROUP_HOVER);
	AddExhaust (th_hover[0], 6, 0.5, _V( 0,-1.6, 3), _V(0,-1,0));
	AddExhaust (th_hover[0], 6, 0.5, _V(-3,-1.3,-3), _V(0,-1,0));
	AddExhaust (th_hover[0], 6, 0.5, _V( 3,-1.3,-3), _V(0,-1,0));
	AddExhaustStream (th_hover[0], _V(0,-4,0), &contrail);
	AddExhaustStream (th_hover[0], _V(0,-2,3), &exhaust_hover);
	AddExhaustStream (th_hover[0], _V(-3,-2,-3), &exhaust_hover);
	AddExhaustStream (th_hover[0], _V( 3,-2,-3), &exhaust_hover);

	// set of attitude thrusters (idealised). The arrangement is such that no angular
	// momentum is created in linear mode, and no linear momentum is created in rotational mode.
	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];
	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 8), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-8), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 8), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-8), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_UP);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_DOWN);
	AddExhaust (th_att_rot[0], 0.6,  0.078, _V(-0.75,-0.7,  9.65), _V(0,-1,0));
	AddExhaust (th_att_rot[0], 0.6,  0.078, _V( 0.75,-0.7,  9.65), _V(0,-1,0));
	AddExhaust (th_att_rot[1], 0.79, 0.103, _V(-0.1 , 0.55,-7.3 ), _V(0, 1,0));
	AddExhaust (th_att_rot[1], 0.79, 0.103, _V( 0.1 , 0.55,-7.3 ), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.6,  0.078, _V(-0.75,-0.3,  9.65), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.6,  0.078, _V( 0.75,-0.3,  9.65), _V(0, 1,0));
	AddExhaust (th_att_rot[3], 0.79, 0.103, _V(-0.1, -0.55,-7.3 ), _V(0,-1,0));
	AddExhaust (th_att_rot[3], 0.79, 0.103, _V( 0.1, -0.55,-7.3 ), _V(0,-1,0));

	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 6), _V(-1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-6), _V( 1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 6), _V( 1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-6), _V(-1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWRIGHT);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_LEFT);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_RIGHT);
	AddExhaust (th_att_rot[0], 0.6,  0.078, _V(0.81,-0.5,9.65), _V(1,0,0));
	AddExhaust (th_att_rot[1], 1.03, 0.134, _V(-2.1,0.3,-5.6), _V(-1,0,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V(-0.81,-0.5,9.65), _V(-1,0,0));
	AddExhaust (th_att_rot[3], 1.03, 0.134, _V(2.1,0.3,-5.6), _V(1,0,0));

	th_att_rot[0] = CreateThruster (_V( 6,0,0), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = CreateThruster (_V(-6,0,0), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = CreateThruster (_V(-6,0,0), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = CreateThruster (_V( 6,0,0), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot, 2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);
	AddExhaust (th_att_rot[0], 1.03, 0.134, _V(-5.1, 0.2,0.4), _V(0, 1,0));
	AddExhaust (th_att_rot[1], 1.03, 0.134, _V( 5.1,-0.8,0.4), _V(0,-1,0));
	AddExhaust (th_att_rot[2], 1.03, 0.134, _V( 5.1, 0.2,0.4), _V(0, 1,0));
	AddExhaust (th_att_rot[3], 1.03, 0.134, _V(-5.1,-0.8,0.4), _V(0,-1,0));

	th_att_lin[0] = CreateThruster (_V(0,0,-7), _V(0,0, 1), 2*MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_lin[1] = CreateThruster (_V(0,0, 7), _V(0,0,-1), 2*MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_lin,   1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_att_lin+1, 1, THGROUP_ATT_BACK);
	AddExhaust (th_att_lin[0], 0.6, 0.078, _V(0,-0.2,-7.6), _V(0,0,-1));
	AddExhaust (th_att_lin[0], 0.6, 0.078, _V(0,0.22,-7.6), _V(0,0,-1));
	AddExhaust (th_att_lin[1], 0.6, 0.078, _V(-0.77,-0.5,9.8), _V(0,0,1));
	AddExhaust (th_att_lin[1], 0.6, 0.078, _V( 0.77,-0.5,9.8), _V(0,0,1));

	// **************************** aerodynamics ***********************************

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-0.2), VLiftCoeff, 5, 100, 1.5);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 5, 15, 1.5);
	// vertical stabiliser and body lift and drag components

	CreateControlSurface (AIRCTRL_ELEVATOR,     1.2, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_XPOS, anim_elevator);
	CreateControlSurface (AIRCTRL_RUDDER,       0.8, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_YPOS, anim_rudder);
	CreateControlSurface (AIRCTRL_AILERON,      0.2, 1.5, _V( 7.5,0,-7.2), AIRCTRL_AXIS_XPOS, anim_raileron);
	CreateControlSurface (AIRCTRL_AILERON,      0.2, 1.5, _V(-7.5,0,-7.2), AIRCTRL_AXIS_XNEG, anim_laileron);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM, 0.1, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_XPOS, anim_elevatortrim);

	CreateVariableDragElement (&gear_proc, 0.8, _V(0, -1, 0));
	CreateVariableDragElement (&nose_proc, 3, _V(0, 0, 8));
	CreateVariableDragElement (&radiator_proc, 1, _V(0,1.5,-4));
	CreateVariableDragElement (&brake_proc, 4, _V(0,0,-8));  // airbrake

	SetRotDrag (_V(0.10,0.13,0.04));
	// angular damping

	// ******************************** mesh ***************************************

	AddMesh (oapiLoadMeshGlobal ("DeltaGlider"));
}

// Define animation sequences for moving parts
void DeltaGlider::DefineAnimations ()
{
	static UINT NosewheelGrp1[4] = {18,23,19,24};
	static ANIMCOMP nosewheel1 = { // nose wheel and front flap
		NosewheelGrp1, 4, 0.3, 0.9,
		0,-1.10,8.553,              // rotation reference
		1,0,0,                      // rotation axis
		(float)(-0.553*PI),         // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NosewheelGrp2[2] = {13,22};
	static ANIMCOMP nosewheel2 = { // nose wheel left side flap
		NosewheelGrp2, 2, 0.0, 0.4,
		-0.3, -1.22046, 7.02812,    // rotation reference
		0,0.045953,0.99894,         // rotation axis
		(float)(0.5*PI),            // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NosewheelGrp3[2] = {12,21};
	static ANIMCOMP nosewheel3 = { // nose wheel right side flap
		NosewheelGrp3, 2, 0.0, 0.42,
		0.3, -1.22046, 7.02812,     // rotation reference
		0,0.045953,0.99894,         // rotation axis
		(float)(-0.5*PI),           // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT RightwheelGrp1[4] = {10,29,30,9};
	static ANIMCOMP rightwheel1 = { // right wheel and flap
		RightwheelGrp1, 4, 0.2, 1.0,
		3.631,-0.799,-4.9499,
		0,0,1,
		(float)(0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT RightwheelGrp2[2] = {16,28};
	static ANIMCOMP rightwheel2 = { // right wheel inner flap
		RightwheelGrp2, 2, 0.04, 0.5,
		1.82,-0.86,-4.4621,
		0,0,1,
		(float)(-0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT LeftwheelGrp1[4] = {15,25,11,26};
	static ANIMCOMP leftwheel1 = { // left wheel and flap
		LeftwheelGrp1, 4, 0.17, 0.96,
		-3.631,-0.799,-4.9499,
		0,0,1,
		(float)(-0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT LeftwheelGrp2[2] = {17,27};
	static ANIMCOMP leftwheel2 = { // left wheel inner flap
		LeftwheelGrp2, 2, 0.02, 0.48,
		-1.82,-0.86,-4.4621,
		0,0,1,
		(float)(0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NoseconeGrp1[2] = {34,35};
	static ANIMCOMP nosecone1 = { // upper right nosecone segment
		NoseconeGrp1, 2, 0.0, 1.0,
		0.601,-0.241,9.837,
		-0.351,0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp2[2] = {38,39};
	static ANIMCOMP nosecone2 = { // lower right nosecone segment
		NoseconeGrp2, 2, 0.0, 1.0,
		0.601,-0.738,9.837,
		0.351,0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp3[2] = {37,40};
	static ANIMCOMP nosecone3 = { // lower left nosecone segment
		NoseconeGrp3, 2, 0.0, 1.0,
		-0.601,-0.738,9.837,
		0.351,-0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp4[2] = {62,36};
	static ANIMCOMP nosecone4 = { // upper left nosecone segment
		NoseconeGrp4, 2, 0.0, 1.0,
		-0.601,-0.241,9.837,
		-0.351,-0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT AirlockGrp1[1] = {56};
	static ANIMCOMP airlock = { // outer airlock
		AirlockGrp1, 1, 0.0, 1.0,
		-0.169,-0.080,9.849,
		1,0,0,
		(float)(0.45*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};

	// Register animation for landing gear
	anim_gear = RegisterAnimSequence (1.0); // gear is fully extended in the mesh
	AddAnimComp (anim_gear, &nosewheel1);   // nose wheel component 1
	AddAnimComp (anim_gear, &nosewheel2);   // nose wheel component 2
	AddAnimComp (anim_gear, &nosewheel3);   // nose wheel component 3
	AddAnimComp (anim_gear, &rightwheel1);  // right wheel component 1
	AddAnimComp (anim_gear, &rightwheel2);  // right wheel component 2
	AddAnimComp (anim_gear, &leftwheel1);   // left wheel component 1
	AddAnimComp (anim_gear, &leftwheel2);   // left wheel component 2

	// Register animation for nose cone
	anim_nose = RegisterAnimSequence (0.0); // nose cone is closed in mesh
	AddAnimComp (anim_nose, &nosecone1);    // nose cone component 1
	AddAnimComp (anim_nose, &nosecone2);    // nose cone component 2
	AddAnimComp (anim_nose, &nosecone3);    // nose cone component 3
	AddAnimComp (anim_nose, &nosecone4);    // nose cone component 4

	// Register animation for airlock
	anim_lock = RegisterAnimSequence (0.0); // airlock is closed in mesh
	AddAnimComp (anim_lock, &airlock);      // airlock component 1

	// ***** Radiator animation *****
	static UINT RaddoorGrp[1] = {73};
	static MGROUP_ROTATE Raddoor (0, RaddoorGrp, 1,
		_V(0,1.481,-3.986), _V(1,0,0), (float)(170*RAD));
	static UINT RadiatorGrp[4] = {74,75,76,77};
	static MGROUP_TRANSLATE Radiator (0, RadiatorGrp, 4,
		_V(0,0.584,-0.157));
	static UINT LRadiatorGrp[1] = {76};
	static MGROUP_ROTATE LRadiator (0, LRadiatorGrp, 1,
		_V(-0.88,1.94,-4.211), _V(0,0.260,0.966), (float)(165*RAD));
	static UINT RRadiatorGrp[1] = {77};
	static MGROUP_ROTATE RRadiator (0, RRadiatorGrp, 1,
		_V(0.93,1.91,-4.211), _V(0,0.260,0.966), (float)(-165*RAD));
	anim_radiator = CreateAnimation (0);
	AddAnimationComponent (anim_radiator, 0, 0.33, &Raddoor);
	AddAnimationComponent (anim_radiator, 0.25, 0.5, &Radiator);
	AddAnimationComponent (anim_radiator, 0.5, 0.75, &LRadiator);
	AddAnimationComponent (anim_radiator, 0.75, 1, &RRadiator);

	// ***** Rudder animation *****
	static UINT RRudderGrp[1] = {67};
	static MGROUP_ROTATE RRudder (0, RRudderGrp, 1,
		_V(8.486,-0.167,-6.005), _V(0.141,0.973,-0.182), (float)(-60*RAD));
	static UINT LRudderGrp[1] = {68};
	static MGROUP_ROTATE LRudder (0, LRudderGrp, 1,
		_V(-8.486, -0.167,-6.005), _V(-0.141,0.973,-0.182), (float)(-60*RAD));
	anim_rudder = CreateAnimation (0.5);
	AddAnimationComponent (anim_rudder, 0, 1, &RRudder);
	AddAnimationComponent (anim_rudder, 0, 1, &LRudder);

	// ***** Elevator animation *****
	static UINT ElevatorGrp[4] = {69,70,71,72};
	static MGROUP_ROTATE Elevator (0, ElevatorGrp, 4,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(40*RAD));
	anim_elevator = CreateAnimation (0.5);
	AddAnimationComponent (anim_elevator, 0, 1, &Elevator);

	// ***** Elevator trim animation *****
	static MGROUP_ROTATE ElevatorTrim (0, ElevatorGrp, 4,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(3.5*RAD));
	anim_elevatortrim = CreateAnimation (0.5);
	AddAnimationComponent (anim_elevatortrim, 0, 1, &ElevatorTrim);

	// ***** Aileron animation *****
	static UINT LAileronGrp[2] = {70,71};
	static MGROUP_ROTATE LAileron (0, LAileronGrp, 2,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(-20*RAD));
	anim_laileron = CreateAnimation (0.5);
	AddAnimationComponent (anim_laileron, 0, 1, &LAileron);

	static UINT RAileronGrp[2] = {69,72};
	static MGROUP_ROTATE RAileron (0, RAileronGrp, 2,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(20*RAD));
	anim_raileron = CreateAnimation (0.5);
	AddAnimationComponent (anim_raileron, 0, 1, &RAileron);

	// ***** Airbrake animation *****
	static UINT UpperBrakeGrp[2] = {70,72};
	static MGROUP_ROTATE UpperBrake (0, UpperBrakeGrp, 2,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(25*RAD));
	static UINT LowerBrakeGrp[2] = {69,71};
	static MGROUP_ROTATE LowerBrake (0, LowerBrakeGrp, 2,
		_V(0,-0.4,-6.0), _V(1,0,0), (float)(-25*RAD));
	static MGROUP_ROTATE RRudderBrake (0, RRudderGrp, 1,
		_V(8.486,-0.167,-6.005), _V(0.141,0.973,-0.182), (float)(25*RAD));
	static MGROUP_ROTATE LRudderBrake (0, LRudderGrp, 1,
		_V(-8.486, -0.167,-6.005), _V(-0.141,0.973,-0.182), (float)(-25*RAD));

	anim_brake = CreateAnimation (0);
	AddAnimationComponent (anim_brake, 0, 1, &UpperBrake);
	AddAnimationComponent (anim_brake, 0, 1, &LowerBrake);
	AddAnimationComponent (anim_brake, 0, 1, &RRudderBrake);
	AddAnimationComponent (anim_brake, 0, 1, &LRudderBrake);
}

// Read status from scenario file
void DeltaGlider::LoadState (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "NOSECONE", 8)) {
			sscanf (line+8, "%d%lf", &nose_status, &nose_proc);
		} else if (!strnicmp (line, "GEAR", 4)) {
			sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
			SetGearParameters (gear_proc);
		} else if (!strnicmp (line, "AIRLOCK", 7)) {
			sscanf (line+7, "%d%lf", &lock_status, &lock_proc);
        } else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
    }

	// update animation states
	SetAnimState (anim_gear, gear_proc);
	SetAnimState (anim_nose, nose_proc);
	SetAnimState (anim_lock, lock_proc);

	SetAnimation (anim_brake, brake_proc);
}

// Write status to scenario file
void DeltaGlider::SaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	sprintf (cbuf, "%d %0.4f", nose_status, nose_proc);
	oapiWriteScenario_string (scn, "NOSECONE", cbuf);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);

	sprintf (cbuf, "%d %0.4f", lock_status, lock_proc);
	oapiWriteScenario_string (scn, "AIRLOCK", cbuf);
}

void DeltaGlider::ReleaseSurfaces ()
{
	for (int i = 0; i < nsurf; i++)
		if (srf[i]) {
			oapiDestroySurface (srf[i]);
			srf[i] = 0;
		}
}

void DeltaGlider::InitPanel (int panel)
{
	int i;

	switch (panel) {
	case 0: // main panel
		srf[0] = oapiCreateSurface (LOADBMP (IDB_DIAL1));
		srf[1] = oapiCreateSurface (LOADBMP (IDB_NAVBUTTON));
		srf[2] = oapiCreateSurface (LOADBMP (IDB_LIGHT1));
		srf[3] = oapiCreateSurface (LOADBMP (IDB_HORIZON));   oapiSetSurfaceColourKey (srf[3], 0);
		srf[4] = oapiCreateSurface (LOADBMP (IDB_SLIDER1));
		srf[5] = oapiCreateSurface (LOADBMP (IDB_SWITCH3));

		for (i = 0; i < 3; i++) engsliderpos[i] = (UINT)-1;
		elevtrimpos = (UINT)-1;
		for (i = 0; i < 2; i++) {
			pg_pos[i] = yg_pos[i] = 0;
			pgimbal[i] = ygimbal[i] = 0.0;
		}
		break;
	case 1: // overhead panel
		srf[0] = oapiCreateSurface (LOADBMP (IDB_SWITCH1));
		srf[1] = oapiCreateSurface (LOADBMP (IDB_INDICATOR));
		srf[2] = oapiCreateSurface (LOADBMP (IDB_VPITCH));
		srf[3] = oapiCreateSurface (LOADBMP (IDB_VBANK));
		srf[4] = oapiCreateSurface (LOADBMP (IDB_VYAW));
		break;
	case 2: // bottom panel
		srf[0] = oapiCreateSurface (LOADBMP (IDB_SWITCH2));
		srf[1] = oapiCreateSurface (LOADBMP (IDB_LEVER1));
		srf[2] = oapiCreateSurface (LOADBMP (IDB_INDICATOR));
		for (i = 0; i < 2; i++) wbrake_pos[i] = (UINT)-1;
		break;
	}
}

// Frame update
void DeltaGlider::Timestep (double simt)
{
	VECTOR3 v;
	double hspd, vspd;
	if (GetHorizonAirspeedVector (v)) {
		hspd = hypot (v.x, v.z);
		vspd = v.y;
		slope = (hspd ? atan (vspd/hspd) : 0.0);
	} else slope = 0.0;
	if ((alt = GetAltitude()) > 0.0 && slope < 0.0)
		range = -alt*hspd/vspd;
	else range = -1.0;

	// yaw gimbal auto adjustment
	if (yg_mode == 1 && yg_comp == 1) {
		double t0, t1, t01;
		t0 = GetThrusterLevel (th_main[0]);
		t1 = GetThrusterLevel (th_main[1]);
		if (t01 = t0+t1)
			SetGimbalYaw (2, (t0-t1)/(t01*7.7));
	}

	// animate landing gear
	if (gear_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
		if (gear_status == DOOR_CLOSING) {
			if (gear_proc > 0.0)
				gear_proc = max (0.0, gear_proc-da);
			else {
				gear_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (2, AID_GEARINDICATOR);
			}
		} else  { // door opening
			if (gear_proc < 1.0)
				gear_proc = min (1.0, gear_proc+da);
			else {
				gear_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (2, AID_GEARINDICATOR);
			}
		}
		SetAnimState (anim_gear, gear_proc);
		SetGearParameters (gear_proc);
	}

	// animate nose cone
	if (nose_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * NOSE_OPERATING_SPEED;
		if (nose_status == DOOR_CLOSING) {
			if (nose_proc > 0.0)
				nose_proc = max (0.0, nose_proc-da);
			else {
				nose_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (2, AID_NOSECONEINDICATOR);
			}
		} else { // door opening
			if (nose_proc < 1.0)
				nose_proc = min (1.0, nose_proc+da);
			else {
				nose_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (2, AID_NOSECONEINDICATOR);
			}
		}
		SetAnimState (anim_nose, nose_proc);
	}

	// animate airlock
	if (lock_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * AIRLOCK_OPERATING_SPEED;
		if (lock_status == DOOR_CLOSING) {
			if (lock_proc > 0.0)
				lock_proc = max (0.0, lock_proc-da);
			else {
				lock_status = DOOR_CLOSED;
				//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
			}
		} else { // door opening
			if (lock_proc < 1.0)
				lock_proc = min (1.0, lock_proc+da);
			else {
				lock_status = DOOR_OPEN;
				//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
			}
		}
		SetAnimState (anim_lock, lock_proc);
	}

	// animate radiator
	if (radiator_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * RADIATOR_OPERATING_SPEED;
		if (radiator_status == DOOR_CLOSING) { // retract radiator
			if (radiator_proc > 0.0) radiator_proc = max (0.0, radiator_proc-da);
			else                     radiator_status = DOOR_CLOSED;
		} else {                               // deploy radiator
			if (radiator_proc < 1.0) radiator_proc = min (1.0, radiator_proc+da);
			else                     radiator_status = DOOR_OPEN;
		}
		SetAnimation (anim_radiator, radiator_proc);
	}

	// animate airbrake
	if (brake_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * AIRBRAKE_OPERATING_SPEED;
		if (brake_status == DOOR_CLOSING) { // retract brake
			if (brake_proc > 0.0) brake_proc = max (0.0, brake_proc-da);
			else                  brake_status = DOOR_CLOSED;
		} else {                            // deploy brake
			if (brake_proc < 1.0) brake_proc = min (1.0, brake_proc+da);
			else                  brake_status = DOOR_OPEN;
		}
		SetAnimation (anim_brake, brake_proc);
	}
}

void DeltaGlider::ActivateLandingGear (DoorStatus action)
{
	if (action == DOOR_OPENING && GroundContact()) return;
	// we cannot deploy the landing gear if we are already sitting on the ground

	gear_status = action;
	oapiTriggerPanelRedrawArea (2, AID_GEARLEVER);
	oapiTriggerPanelRedrawArea (2, AID_GEARINDICATOR);
}

void DeltaGlider::RevertLandingGear ()
{
	ActivateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::SetGearParameters (double state)
{
	if (state == 1.0) {
		SetTouchdownPoints (_V(0,-2.57,10), _V(-3.5,-2.57,-3), _V(3.5,-2.57,-3));
		SetSurfaceFrictionCoeff (0.05, 0.4);
	} else {
		SetTouchdownPoints (_V(0,-1.5,9), _V(-6,-0.8,-5), _V(3,-1.2,-5));
		SetSurfaceFrictionCoeff (0.4, 0.4);
	}
}

void DeltaGlider::ActivateDockingPort (DoorStatus action)
{
	nose_status = action;
	oapiTriggerPanelRedrawArea (2, AID_NOSECONELEVER);
	oapiTriggerPanelRedrawArea (2, AID_NOSECONEINDICATOR);
}

void DeltaGlider::RevertDockingPort ()
{
	ActivateDockingPort (nose_status == DOOR_CLOSED || nose_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::ActivateAirlock (DoorStatus action)
{
	lock_status = action;
	oapiTriggerPanelRedrawArea (1, AID_OUTERAIRLOCKBUTTON);
	//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
}

void DeltaGlider::RevertAirlock ()
{
	ActivateAirlock (lock_status == DOOR_CLOSED || lock_status == DOOR_CLOSING ?
		             DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::ActivateAirbrake (DoorStatus action)
{
	brake_status = action;
}

void DeltaGlider::RevertAirbrake (void)
{
	ActivateAirbrake (brake_status == DOOR_CLOSED || brake_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::ActivateRadiator (DoorStatus action)
{
	radiator_status = action;
}

void DeltaGlider::RevertRadiator (void)
{
	ActivateRadiator (radiator_status == DOOR_CLOSED || radiator_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

bool DeltaGlider::DecAttMode ()
{
	int mode = GetAttitudeMode();
	if (mode) {
		SetAttitudeMode (mode-1);
		return true;
	} else return false;
}

bool DeltaGlider::IncAttMode ()
{
	int mode = GetAttitudeMode();
	if (mode < 2) {
		SetAttitudeMode (mode+1);
		return true;
	} else return false;
}

bool DeltaGlider::DecADCMode ()
{
	DWORD mode = min (GetADCtrlMode(),2);
	if (mode) {
		SetADCtrlMode (mode-1);
		return true;
	} else return false;
}

bool DeltaGlider::IncADCMode ()
{
	DWORD mode = GetADCtrlMode();
	if (mode <= 1) {
		SetADCtrlMode (mode ? 7 : 1);
		return true;
	} else return false;
}

void DeltaGlider::SetMainRetroLevel (int which, double lmain, double lretro)
{
	if (which == 2) {   // set main thruster group
		SetThrusterGroupLevel (THGROUP_MAIN,  lmain);
		SetThrusterGroupLevel (THGROUP_RETRO, lretro);
	} else {            // set individual engine
		SetThrusterLevel (th_main [which], lmain);
		SetThrusterLevel (th_retro[which], lretro);
	}
}

void DeltaGlider::SetGimbalPitch (int which, int pos)
{
	int i1 = (which == 1 ? 1:0);
	int i2 = (which == 0 ? 1:2);
	int i, p;

	for (i = i1; i < i2; i++) {
		// make the slider stick in center position
		if (pos <= -4)     pg_pos[i] = pos, p = pos+4;
		else if (pos >= 4) pg_pos[i] = pos, p = pos-4;
		else               pg_pos[i] = p = 0;

		pgimbal[i] = p/22.0 * RAD*1.0;
		// make pitch gimbal range += 1.0
		SetThrusterDir (th_main[i], _V(ygimbal[i], pgimbal[i], 1.0));
	}
	oapiTriggerPanelRedrawArea (0, AID_PITCHGIMBAL);
}

void DeltaGlider::SetGimbalYaw (int which, double tan_gmb)
{
	// make the slider stick in center position
	int slpos;
	if      (!tan_gmb)    slpos = 0;
	else if (tan_gmb > 0) slpos = -(int)(tan_gmb*7.7*22.0)-4;
	else                  slpos = -(int)(tan_gmb*7.7*22.0)+4;

	if (which != 1) SetThrusterDir (th_main[0], _V(ygimbal[0] = tan_gmb, pgimbal[0], 1.0)), yg_pos[0] = slpos;
	if (which != 0) SetThrusterDir (th_main[1], _V(ygimbal[1] = tan_gmb, pgimbal[1], 1.0)), yg_pos[1] = slpos;

	oapiTriggerPanelRedrawArea (0, AID_YAWGIMBAL);
}

bool DeltaGlider::SwitchGimbalYawMode (bool switchup)
{
	if (switchup) {
		if (yg_mode == 2) return false;
		yg_mode++;
	} else {
		if (!yg_mode) return false;
		yg_mode--;
	}
	if      (yg_mode == 0)   SetGimbalYaw (2, 0.0);  // neutral position
	else if (yg_mode == 1 && yg_comp == 0) { // align both with CG
		SetGimbalYaw (0,  1.0/7.7);
		SetGimbalYaw (1, -1.0/7.7);
	}
	return true;
}

bool DeltaGlider::SwitchGimbalYawComp (bool switchup)
{
	if (switchup == (yg_comp > 0)) return false;
	yg_comp = 1-yg_comp;
	if (yg_mode == 1 && yg_comp == 0) {
		SetGimbalYaw (0,  1.0/7.7);
		SetGimbalYaw (1, -1.0/7.7);
	}
	return true;
}

bool DeltaGlider::SwitchGimbalPitchMode (bool switchup)
{
	if (switchup == (pg_mode > 0)) return false;
	pg_mode = 1-pg_mode;
	if (pg_mode == 0) SetGimbalPitch (2, 0);
	return true;
}

void DeltaGlider::RedrawPanel_PropellantStatus (SURFHANDLE surf)
{
	char cbuf[16];
	int i;
	double m, lvl;

	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	SetTextColor (hDC, RGB(0, 255, 0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	SelectObject (hDC, g_Param.brush[0]);
	SelectObject (hDC, GetStockObject (NULL_PEN));

	// main tank status
	m    = GetPropellantMass (ph_main);
	lvl  = m/MAX_MAIN_FUEL;
	if (i = (int)(101.0*lvl))
		Rectangle (hDC, 16, 0, i+17, 8);
	sprintf (cbuf, "%0.1f", 100.0*lvl);
	TextOut (hDC, 40, 20, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.0f", m);
	TextOut (hDC, 40, 41, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", GetPropellantFlowrate (ph_main));
	TextOut (hDC, 40, 62, cbuf, strlen(cbuf));

	// RCS tank status
	m  = GetPropellantMass (ph_rcs);
	lvl  = m/MAX_RCS_FUEL;
	if (i = (int)(101.0*lvl))
		Rectangle (hDC, 16, 8, i+17, 16);
	sprintf (cbuf, "%0.1f", 100.0*lvl);
	TextOut (hDC, 112, 20, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", m);
	TextOut (hDC, 112, 41, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.3f", GetPropellantFlowrate (ph_rcs));
	TextOut (hDC, 112, 62, cbuf, strlen(cbuf));

	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_EngineStatus (SURFHANDLE surf)
{
	char cbuf[16];
	double lvlm, lvlr = 0.0, lvlh, FM, FH;
	double m = GetMass();
	int scl;
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	SelectObject (hDC, g_Param.brush[1]);
	SelectObject (hDC, GetStockObject (NULL_PEN));

	if (lvlm = GetThrusterGroupLevel (thg_main)) {
		scl = (int)(lvlm*82.0);
		Rectangle (hDC, 27, 82, 37, 82-scl);
	} else if (lvlr = GetThrusterGroupLevel (thg_retro)) {
		scl = (int)(lvlr*22.0);
		Rectangle (hDC, 38, 80, 48, 80+scl);
	}
	if (lvlh = GetThrusterGroupLevel (thg_hover)) {
		scl = (int)(lvlh*102.0);
		Rectangle (hDC, 76, 102, 86, 102-scl);
	}
	FM = 2.0 * (lvlm ? lvlm*MAX_MAIN_THRUST : -lvlr*MAX_RETRO_THRUST);
	FH = lvlh*MAX_HOVER_THRUST;
	SetTextColor (hDC, RGB(0,255,0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	sprintf (cbuf, "%0.1f", (lvlm ? lvlm : -lvlr)*100.0);
	TextOut (hDC, 39, 108, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", lvlh*100.0);
	TextOut (hDC, 106, 108, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", FM*1e-3);
	TextOut (hDC, 39, 129, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", FH*1e-3);
	TextOut (hDC, 106, 129, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", FM/m);
	TextOut (hDC, 39, 150, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", FH/m);
	TextOut (hDC, 106, 150, cbuf, strlen(cbuf));
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_AOA (SURFHANDLE surf)
{
	static const double dial_max = RAD*165.0;
	double dial_angle = 7.7 * GetAOA();
	if (fabs (dial_angle) > dial_max)
		dial_angle = (dial_angle > 0.0 ? dial_max:-dial_max);
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 28, 28, 26.0, PI-dial_angle, &aoa_ind);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Slip (SURFHANDLE surf)
{
	static const double dial_max = RAD*165.0;
	double dial_angle = 7.7 * GetSlipAngle();
	if (fabs (dial_angle) > dial_max)
		dial_angle = (dial_angle > 0.0 ? dial_max:-dial_max);
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 28, 28, 26.0, PI*0.5-dial_angle, &slip_ind);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Slope (SURFHANDLE surf)
{
	const double smax = 7.0*RAD, smin = -22.0*RAD;
	double alpha;
	if      (slope > smax) alpha = smax;
	else if (slope < smin) alpha = smin;
	else                   alpha = slope;
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 140, 19, 140.0, PI-alpha, &slope_ind, 0.2);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Range (SURFHANDLE surf)
{
	char cbuf[16];
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	SetTextColor (hDC, RGB(255, 0, 0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	sprintf (cbuf, "%+0.1f", DEG*slope);
	TextOut (hDC, 34, 0, cbuf, strlen(cbuf));
	if (range > 0.0 && range < 1e7) {
		sprintf (cbuf, "%0.0f", 0.001*range);
		TextOut (hDC, 34, 30, cbuf, strlen(cbuf));
	} else
		TextOut (hDC, 34, 30, "N/A", 3);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Navmode (SURFHANDLE surf)
{
	for (DWORD i = NAVMODE_KILLROT; i <= NAVMODE_HOLDALT; i++)
		if (GetNavmodeState (i))
				oapiBlt (surf, srf[1], ((i-1)/2) * 39, ((i-1)%2) * 39, (i-1)*37, 0, 37, 37);
}

bool DeltaGlider::RedrawPanel_Throttle (SURFHANDLE surf)
{
	UINT i, pos;
	bool redraw = false;

	for (i = 0; i < 2; i++) {
		double level = GetThrusterLevel (th_main[i]);
		if (level > 0) pos = (UINT)((1.0-level)*108.0);
		else {
			level = GetThrusterLevel (th_retro[i]);
			if (level > 0) pos = 125+(UINT)(level*32.0);
			else           pos = 116;
		}
		if (pos != engsliderpos[i])
			engsliderpos[i] = pos, redraw = true;
	}
	if (redraw)
		for (i = 0; i < 2; i++)
			oapiBlt (surf, srf[4], i*26, engsliderpos[i], 0, 0, 23, 18);
	return redraw;
}

bool DeltaGlider::RedrawPanel_Hover (SURFHANDLE surf)
{
	bool redraw = false;
	double level = GetThrusterLevel (th_hover[0]);
	UINT pos = (UINT)((1.0-level)*116.0);
	if (pos != engsliderpos[2]) {
		oapiBlt (surf, srf[4], 0, engsliderpos[2] = pos, 0, 0, 23, 18);
		redraw = true;
	}
	return redraw;
}

bool DeltaGlider::RedrawPanel_ElevTrim (SURFHANDLE surf)
{
	double level = GetControlSurfaceLevel (AIRCTRL_ELEVATORTRIM);
	UINT pos = (UINT)((1.0+level)*23.0);
	if (pos != elevtrimpos) {
		HDC hDC = oapiGetDC (surf);
		SelectObject (hDC, g_Param.brush[4]);
		SelectObject (hDC, GetStockObject (NULL_PEN));
		Rectangle (hDC, 1, pos, 17, pos+7);
		oapiReleaseDC (surf, hDC);
		elevtrimpos = pos;
		return true;
	} else return false;
}

void DeltaGlider::RedrawPanel_PGimbal (SURFHANDLE surf)
{
	for (int i = 0; i < 2; i++)
		oapiBlt (surf, srf[4], i*20, pg_pos[i]+26, 0, 19, 18, 9);
}

void DeltaGlider::RedrawPanel_PYaw (SURFHANDLE surf)
{
	for (int i = 0; i < 2; i++)
		oapiBlt (surf, srf[4], yg_pos[i]+26, i*20, 23, 0, 9, 18);
}

bool DeltaGlider::RedrawPanel_RotVelocity (SURFHANDLE surf, int which)
{
	int idx;
	double v, av;
	VECTOR3 vrot;
	GetAngularVel (vrot);
	v  = (which == AID_VPITCH ? -vrot.x : which == AID_VBANK ? -vrot.z : vrot.y);
	av = fabs(v*DEG);

	if      (av <  1.0) idx = 0;
	else if (av < 11.0) idx = 1 + (int)((av-1.0)*0.4);
	else if (av < 45.0) idx = 5 + (int)((av-11.0)*3.0/34.0);
	else                idx = 8;
	if (v >= 0.0) idx  = 8-idx;
	else          idx += 8;

	switch (which) {
		case AID_VPITCH: oapiBlt (surf, srf[2], 0, 0, idx*40, 0, 40, 49); break;
		case AID_VBANK:  oapiBlt (surf, srf[3], 0, 0, idx*50, 0, 50, 40); break;
		case AID_VYAW:   oapiBlt (surf, srf[4], 0, 0, idx*50, 0, 50, 40); break;
	}
	return true;
}

bool DeltaGlider::RedrawPanel_RotAcceleration (SURFHANDLE surf, int which)
{
	int idx;
	double a, aa;
	VECTOR3 arot;
	GetAngularAcc (arot);
	a  = (which == AID_APITCH ? -arot.x : which == AID_ABANK ? -arot.z : arot.y);
	a *= 2.0;
	aa = fabs(a*DEG);

	if      (aa <  1.0) idx = 0;
	else if (aa < 11.0) idx = 1 + (int)((aa-1.0)*0.4);
	else if (aa < 45.0) idx = 5 + (int)((aa-11.0)*3.0/34.0);
	else                idx = 8;
	if (a >= 0.0) idx  = 8-idx;
	else          idx += 8;

	switch (which) {
		case AID_APITCH: oapiBlt (surf, srf[2], 0, 0, idx*40, 0, 40, 49); break;
		case AID_ABANK:  oapiBlt (surf, srf[3], 0, 0, idx*50, 0, 50, 40); break;
		case AID_AYAW:   oapiBlt (surf, srf[4], 0, 0, idx*50, 0, 50, 40); break;
	}
	return true;
}

bool DeltaGlider::RedrawPanel_RotMoment (SURFHANDLE surf, int which)
{
	int idx;
	double m, am;
	VECTOR3 amom;
	GetAngularMoment (amom);
	m  = (which == AID_MPITCH ? -amom.x : which == AID_MBANK ? -amom.z : amom.y);
	m *= 1e-3;
	am = fabs(m);

	if      (am <  1.0) idx = 0;
	else if (am < 11.0) idx = 1 + (int)((am-1.0)*0.4);
	else if (am < 45.0) idx = 5 + (int)((am-11.0)*3.0/34.0);
	else                idx = 8;
	if (m >= 0.0) idx  = 8-idx;
	else          idx += 8;

	switch (which) {
		case AID_MPITCH: oapiBlt (surf, srf[2], 0, 0, idx*40, 0, 40, 49); break;
		case AID_MBANK:  oapiBlt (surf, srf[3], 0, 0, idx*50, 0, 50, 40); break;
		case AID_MYAW:   oapiBlt (surf, srf[4], 0, 0, idx*50, 0, 50, 40); break;
	}
	return true;
}

bool DeltaGlider::RedrawPanel_WBrake (SURFHANDLE surf, int which)
{
	double level = GetWheelbrakeLevel (which);
	UINT pos = (UINT)(level*93.0);
	if (pos != wbrake_pos[which-1]) {
		oapiBlt (surf, srf[0], 0, pos, (which-1)*34, 0, 34, 86);
		wbrake_pos[which-1] = pos;
		return true;
	}
	return false;
}

void DeltaGlider::RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side)
{
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[1]);
	SetTextColor (hDC, RGB(196, 196, 196));
	SetTextAlign (hDC, TA_CENTER);
	SetBkMode (hDC, TRANSPARENT);
	const char *label;
	for (int bt = 0; bt < 6; bt++) {
		if (label = oapiMFDButtonLabel (mfd, bt+side*6))
			TextOut (hDC, 13, 3+41*bt, label, strlen(label));
		else break;
	}
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::DrawNeedle (HDC hDC, int x, int y, double rad, double angle, double *pangle, double vdial)
{
	if (pangle) { // needle response delay
		double dt = oapiGetSimStep();
		if (fabs (angle - *pangle)/dt >= vdial)
			angle = (angle > *pangle ? *pangle+vdial*dt : *pangle-vdial*dt);
		*pangle = angle;
	}
	double dx = rad * cos(angle), dy = rad * sin(angle);
	SelectObject (hDC, g_Param.pen[1]);
	MoveToEx (hDC, x, y, 0); LineTo (hDC, x + (int)(0.85*dx+0.5), y - (int)(0.85*dy+0.5));
	SelectObject (hDC, g_Param.pen[0]);
	MoveToEx (hDC, x, y, 0); LineTo (hDC, x + (int)(dx+0.5), y - (int)(dy+0.5));
}

// ==============================================================
// DLL entry point
// ==============================================================

BOOL WINAPI DllMain (HINSTANCE hModule,
					 DWORD ul_reason_for_call,
					 LPVOID lpReserved)
{
	int i;

	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		g_Param.hDLL = hModule;
		// allocate GDI resources
		g_Param.font[0]  = CreateFont (-13, 0, 0, 0, 700, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
		g_Param.font[1]  = CreateFont (-10, 0, 0, 0, 400, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
		g_Param.brush[0] = CreateSolidBrush (RGB(0,255,0));    // green
		g_Param.brush[1] = CreateSolidBrush (RGB(255,0,0));    // red
		g_Param.brush[2] = CreateSolidBrush (RGB(80,80,224));  // blue
		g_Param.brush[3] = CreateSolidBrush (RGB(160,120,64));  // brown
		g_Param.brush[4] = CreateSolidBrush (RGB(210,210,210)); // white
		g_Param.pen[0] = CreatePen (PS_SOLID, 1, RGB(224,224,224));
		g_Param.pen[1] = CreatePen (PS_SOLID, 3, RGB(164,164,164));
		break;
	case DLL_PROCESS_DETACH:
		// deallocate GDI resources
		for (i = 0; i < 2; i++) DeleteObject (g_Param.font[i]);
		for (i = 0; i < 5; i++) DeleteObject (g_Param.brush[i]);
		for (i = 0; i < 2; i++) DeleteObject (g_Param.pen[i]);
		break;
	}
	return TRUE;
}

// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	// need to init device-dependent resources here in case the screen mode has changed
	g_Param.col[2] = oapiGetColour(80,80,224);
	g_Param.col[3] = oapiGetColour(160,120,64);

	return new DeltaGlider (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (DeltaGlider*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	((DeltaGlider*)vessel)->SetClassCaps (cfg);
}

// Read status from scenario file
DLLCLBK void ovcLoadStateEx (VESSEL *vessel, FILEHANDLE scn, void *vs)
{
	((DeltaGlider*)vessel)->LoadState (scn, vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((DeltaGlider*)vessel)->SaveState (scn);
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((DeltaGlider*)vessel)->Timestep (simt);
}

DLLCLBK void ovcRCSmode (VESSEL *vessel, int mode)
{
	oapiTriggerPanelRedrawArea (0, AID_ATTITUDEMODE);
}

DLLCLBK void ovcADCtrlmode (VESSEL *vessel, DWORD mode)
{
	oapiTriggerPanelRedrawArea (0, AID_ADCTRLMODE);
}

DLLCLBK void ovcNavmode (VESSEL *vessel, int mode, bool active)
{
	oapiTriggerPanelRedrawArea (0, AID_NAVMODE);
}

DLLCLBK void ovcHUDmode (VESSEL *vessel, int mode)
{
	oapiTriggerPanelRedrawArea (0, AID_HUDMODE);
}

DLLCLBK void ovcMFDmode (VESSEL *vessel, int mfd, int mode)
{
	switch (mfd) {
	case MFD_LEFT:
		oapiTriggerPanelRedrawArea (0, AID_MFD1_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD1_RBUTTONS);
		break;
	case MFD_RIGHT:
		oapiTriggerPanelRedrawArea (0, AID_MFD2_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD2_RBUTTONS);
		break;
	}
}

DLLCLBK bool ovcLoadPanel (VESSEL *vessel, int id)
{
	int xofs;
	DeltaGlider *dg = (DeltaGlider*)vessel;
	dg->ReleaseSurfaces();

	HBITMAP hBmp = LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (IDB_PANEL1+id));
	MFDSPEC mfds_left  = {{227, 61, 517, 351}, 6, 6, 47, 41};
	MFDSPEC mfds_right = {{636, 61, 926, 351}, 6, 6, 47, 41};

	switch (id) {
	case 0: // main panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_BOTTOM|PANEL_MOVEOUT_BOTTOM, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,1,2);
		oapiRegisterMFD (MFD_LEFT,  mfds_left);
		oapiRegisterMFD (MFD_RIGHT, mfds_right);
		oapiRegisterPanelArea (AID_PROPELLANTSTATUS,  _R(  27,  81,  144, 156), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINESTATUS,      _R(1000,  90, 1112, 256), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ATTITUDEMODE,      _R( 107, 195,  147, 239), PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiRegisterPanelArea (AID_ADCTRLMODE,        _R(  23, 195,   63, 239), PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiRegisterPanelArea (AID_NAVMODE,           _R(  38, 247,  153, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_HUDMODE,           _R(  15,  18,  122,  33), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_AOAINSTR,          _R( 206, 413,  262, 469), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_SLIPINSTR,         _R( 305, 413,  361, 469), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_SLOPEINSTR,        _R( 414, 405,  455, 476), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_RANGEINSTR,        _R( 476, 422,  510, 467), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_HORIZON,           _R( 569, 394,  665, 490), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);

		oapiRegisterPanelArea (AID_ENGINEMAIN,        _R(1001, 272, 1050, 447), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBDOWN | PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINEHOVER,       _R(1083, 272, 1106, 406), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_PITCHGIMBAL,       _R( 108, 371,  146, 432), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN | PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_YAWGIMBAL,         _R(  28, 383,   89, 421), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN | PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_YAWGIMBALMODE,     _R(  31, 455,   56, 493), PANEL_REDRAW_MOUSE, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_YAWGIMBALCOMP,     _R(  60, 455,   85, 493), PANEL_REDRAW_MOUSE, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_PITCHGIMBALMODE,   _R( 115, 455,  140, 493), PANEL_REDRAW_MOUSE, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_ELEVATORTRIM,      _R(1096, 432, 1112, 484), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);

		xofs = 182; // left MFD
		oapiRegisterPanelArea (AID_MFD1_BBUTTONS,     _R( 55+xofs,359,325+xofs, 377), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN); // bottom button row
		oapiRegisterPanelArea (AID_MFD1_LBUTTONS,     _R(    xofs,100, 25+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // left button column
		oapiRegisterPanelArea (AID_MFD1_RBUTTONS,     _R(353+xofs,100,378+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // right button column

		xofs = 591; // right MFD
		oapiRegisterPanelArea (AID_MFD2_BBUTTONS,     _R( 55+xofs,359,325+xofs, 377), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN); // bottom button row
		oapiRegisterPanelArea (AID_MFD2_LBUTTONS,     _R(    xofs,100, 25+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // left button column
		oapiRegisterPanelArea (AID_MFD2_RBUTTONS,     _R(353+xofs,100,378+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // right button column
		break;

	case 1: // overhead panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_TOP|PANEL_MOVEOUT_TOP, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,-1,0);
		oapiRegisterPanelArea (AID_OUTERAIRLOCKBUTTON,_R( 240,  30, 265,  68), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_VPITCH,            _R( 625,  22, 665,  71), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_VBANK,             _R( 557,  22, 607,  62), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_VYAW,              _R( 491,  22, 541,  62), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_APITCH,            _R( 625,  95, 665, 144), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_ABANK,             _R( 557,  95, 607, 135), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_AYAW,              _R( 491,  95, 541, 135), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_MPITCH,            _R( 625, 168, 665, 217), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE); 
		oapiRegisterPanelArea (AID_MBANK,             _R( 557, 168, 607, 208), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		oapiRegisterPanelArea (AID_MYAW,              _R( 491, 168, 541, 208), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);
		break;

	case 2: // bottom panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_TOP|PANEL_ATTACH_BOTTOM, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,0,-1);
		oapiRegisterPanelArea (AID_WBRAKE_BOTH,       _R( 250, 377, 264, 557), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBPRESSED);
		oapiRegisterPanelArea (AID_WBRAKE_LEFT,       _R( 221, 377, 255, 557), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_WBRAKE_RIGHT,      _R( 259, 377, 293, 557), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_GEARLEVER,         _R( 357, 354, 389, 508), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_NOSECONELEVER,     _R( 432, 354, 464, 508), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_GEARINDICATOR,     _R( 328, 509, 357, 540), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_NOSECONEINDICATOR, _R( 465, 509, 494, 540), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		break;
	}
	dg->InitPanel (id);
	return hBmp != NULL;
}

DLLCLBK bool ovcPanelRedrawEvent (VESSEL *vessel, int id, int event, SURFHANDLE surf)
{
	DeltaGlider *dg = (DeltaGlider*)vessel;
	int i;

	switch (id) {
	// panel 0 events:
	case AID_PROPELLANTSTATUS:
		dg->RedrawPanel_PropellantStatus (surf);
		return true;
	case AID_ENGINESTATUS:
		dg->RedrawPanel_EngineStatus (surf);
		return true;
	case AID_ATTITUDEMODE:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->GetAttitudeMode()*40, 0, 40, 44);
		return true;
	case AID_ADCTRLMODE:
		oapiBlt (surf, dg->srf[0], 0, 0, min(dg->GetADCtrlMode(),2)*40, 0, 40, 44);
		return true;
	case AID_NAVMODE:
		dg->RedrawPanel_Navmode (surf);
		return true;
	case AID_HUDMODE:
		if ((i = oapiGetHUDMode()) > HUD_NONE)
			oapiBlt (surf, dg->srf[2], i*29+6, 0, 7, 0, 7, 7);
		return true;
	case AID_AOAINSTR:
		dg->RedrawPanel_AOA (surf);
		return true;
	case AID_SLIPINSTR:
		dg->RedrawPanel_Slip (surf);
		return true;
	case AID_SLOPEINSTR:
		dg->RedrawPanel_Slope (surf);
		return true;
	case AID_RANGEINSTR:
		dg->RedrawPanel_Range (surf);
		return true;
	case AID_HORIZON:
		dg->RedrawPanel_Horizon (surf);
		return true;
	case AID_ENGINEMAIN:
		return dg->RedrawPanel_Throttle (surf);
	case AID_ENGINEHOVER:
		return dg->RedrawPanel_Hover (surf);
	case AID_PITCHGIMBAL:
		dg->RedrawPanel_PGimbal (surf);
		return true;
	case AID_YAWGIMBAL:
		dg->RedrawPanel_PYaw (surf);
		return true;
	case AID_YAWGIMBALMODE:
		oapiBlt (surf, dg->srf[5], 0, 0, dg->yg_mode*25, 0, 25, 38);
		return true;
	case AID_YAWGIMBALCOMP:
		oapiBlt (surf, dg->srf[5], 0, 0, (dg->yg_comp+5)*25, 0, 25, 38);
		return true;
	case AID_PITCHGIMBALMODE:
		oapiBlt (surf, dg->srf[5], 0, 0, (dg->pg_mode+3)*25, 0, 25, 38);
		return true;
	case AID_ELEVATORTRIM:
		return dg->RedrawPanel_ElevTrim (surf);
	case AID_MFD1_LBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_LEFT, 0);
		return true;
	case AID_MFD1_RBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_LEFT, 1);
		return true;
	case AID_MFD2_LBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_RIGHT, 0);
		return true;
	case AID_MFD2_RBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_RIGHT, 1);
		return true;

	case AID_OUTERAIRLOCKBUTTON:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->lock_status == DeltaGlider::DOOR_OPEN ||
			dg->lock_status == DeltaGlider::DOOR_OPENING ? 0 : 25, 0, 25, 38);
		return true;
	case AID_VPITCH:
	case AID_VBANK:
	case AID_VYAW:
		return dg->RedrawPanel_RotVelocity (surf, id);
	case AID_APITCH:
	case AID_ABANK:
	case AID_AYAW:
		return dg->RedrawPanel_RotAcceleration (surf, id);
	case AID_MPITCH:
	case AID_MBANK:
	case AID_MYAW:
		return dg->RedrawPanel_RotMoment (surf, id);

	// panel 2 events
	case AID_WBRAKE_LEFT:
		return dg->RedrawPanel_WBrake (surf, 1);
	case AID_WBRAKE_RIGHT:
		return dg->RedrawPanel_WBrake (surf, 2);
	case AID_GEARLEVER:
		oapiBlt (surf, dg->srf[1], 0, 0, dg->gear_status == DeltaGlider::DOOR_OPEN ||
			dg->gear_status == DeltaGlider::DOOR_OPENING ? 32 : 0, 0, 32, 154);
		return true;
	case AID_NOSECONELEVER:
		oapiBlt (surf, dg->srf[1], 0, 0, dg->nose_status == DeltaGlider::DOOR_OPEN ||
			dg->nose_status == DeltaGlider::DOOR_OPENING ? 96 : 64, 0, 32, 154);
		return true;
	case AID_GEARINDICATOR:
		switch (dg->gear_status) {
		case DeltaGlider::DOOR_CLOSED: oapiBlt (surf, dg->srf[2], 0,  0, 0,  0, 29, 5); break;
		case DeltaGlider::DOOR_OPEN:   oapiBlt (surf, dg->srf[2], 0, 26, 0,  5, 29, 5); break;
		default:                       oapiBlt (surf, dg->srf[2], 0, 13, 0, 20, 29, 5); break;
		}
		return true;
	case AID_NOSECONEINDICATOR:
		switch (dg->nose_status) {
		case DeltaGlider::DOOR_CLOSED: oapiBlt (surf, dg->srf[2], 0,  0, 0, 10, 29, 5); break;
		case DeltaGlider::DOOR_OPEN:   oapiBlt (surf, dg->srf[2], 0, 26, 0, 15, 29, 5); break;
		default:                       oapiBlt (surf, dg->srf[2], 0, 13, 0, 20, 29, 5); break;
		}
		return true;
	}
	return false;
}

DLLCLBK bool ovcPanelMouseEvent (VESSEL *vessel, int id, int event, int mx, int my)
{
	DeltaGlider *dg = (DeltaGlider*)vessel;
	int mfd;
	static int ctrl = 0;

	switch (id) {
	// panel 0 events:
	case AID_ATTITUDEMODE:
		if (event & PANEL_MOUSE_LBDOWN) return dg->DecAttMode();
		if (event & PANEL_MOUSE_RBDOWN) return dg->IncAttMode();
		return false;
	case AID_ADCTRLMODE:
		if (event & PANEL_MOUSE_LBDOWN) return dg->DecADCMode();
		if (event & PANEL_MOUSE_RBDOWN) return dg->IncADCMode();
		return false;
	case AID_NAVMODE:
		dg->ToggleNavmode ((mx/39)*2 + my/39 + 1);
		return true;
	case AID_HUDMODE:
		if (mx%29 < 20) oapiSetHUDMode (HUD_NONE+(mx/29));
		return true;
	case AID_ENGINEMAIN:
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (mx <  12) ctrl = 0; // left engine
			else if (mx >= 37) ctrl = 1; // right engine
			else               ctrl = 2; // both
		}
		if ((my -= 9) < 0) my = 0;
		else if (my > 157) my = 157;
		dg->SetMainRetroLevel (ctrl, my <= 108 ? 1.0-my/108.0  : 0.0,   // main thruster level
			                         my >= 125 ? (my-125)/32.0 : 0.0);  // retro thruster level
		return true;
	case AID_ENGINEHOVER:
		my -= 9;
		if      (my <   0) my =   0;
		else if (my > 116) my = 116;
		dg->SetThrusterGroupLevel (dg->thg_hover, 1.0-my/116.0);
		return true;
	case AID_PITCHGIMBAL:
		if (dg->pg_mode != 1) return false; // not in manual mode
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (mx <  10) ctrl = 0;  // left slider
			else if (mx >= 28) ctrl = 1;  // right slider
			else               ctrl = 2;  // both
		}
		if ((my -= 4) < 0) my = 0;
		else if (my > 52) my = 52;
		dg->SetGimbalPitch (ctrl, my-26);
		return true;
	case AID_YAWGIMBAL:
		if (dg->yg_mode != 2) return false; // not in manual mode
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (my <  10) ctrl = 0;  // left slider
			else if (my >= 28) ctrl = 1; // right slider
			else               ctrl = 2; // both
		}
		if ((mx -= 4) < 0) mx = 0;
		else if (mx > 52) mx = 52;
		// make slider stick in neutral position
		if      (mx <= 22) mx -= 22;
		else if (mx >= 30) mx -= 30;
		else               mx = 0;
		dg->SetGimbalYaw (ctrl, mx / (-22.0*7.7));
		return true;
	case AID_YAWGIMBALMODE:
		return dg->SwitchGimbalYawMode (my < 19);
	case AID_YAWGIMBALCOMP:
		return dg->SwitchGimbalYawComp (my < 19);
	case AID_PITCHGIMBALMODE:
		return dg->SwitchGimbalPitchMode (my < 19);
	case AID_ELEVATORTRIM:
		dg->SetControlSurfaceLevel (AIRCTRL_ELEVATORTRIM,
			dg->GetControlSurfaceLevel (AIRCTRL_ELEVATORTRIM) + 
			oapiGetSimStep() * (my < 22 ? -0.2:0.2));
		return true;
	case AID_MFD1_BBUTTONS:
	case AID_MFD2_BBUTTONS:
		mfd = (id == AID_MFD1_BBUTTONS ? MFD_LEFT : MFD_RIGHT);
		if (mx < 26)                    oapiToggleMFD_on (mfd);
		else if (mx >= 214 && mx < 240) oapiSendMFDKey (mfd, OAPI_KEY_F1);
		else if (mx > 244)              oapiSendMFDKey (mfd, OAPI_KEY_GRAVE);
		return true;
	case AID_MFD1_LBUTTONS:
	case AID_MFD1_RBUTTONS:
	case AID_MFD2_LBUTTONS:
	case AID_MFD2_RBUTTONS:
		if (my%41 < 18) {
			mfd = (id <= AID_MFD1_RBUTTONS ? MFD_LEFT : MFD_RIGHT);
			int bt = my/41 + (id == AID_MFD1_LBUTTONS || id == AID_MFD2_LBUTTONS ? 0 : 6);
			oapiProcessMFDButton (mfd, bt, event);
			return true;
		}
		break;

	// panel 1 events:
	case AID_OUTERAIRLOCKBUTTON:
		dg->ActivateAirlock (my < 19 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;

	// panel 2 events:
	case AID_WBRAKE_BOTH:
	case AID_WBRAKE_LEFT:
	case AID_WBRAKE_RIGHT:
		my -= 43;
		if      (my <  0) my =  0;
		else if (my > 93) my = 93;
		dg->SetWheelbrakeLevel (my/93.0, id-AID_WBRAKE_BOTH, true);
		return true;
	case AID_GEARLEVER:
		dg->ActivateLandingGear (my < 77 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;
	case AID_NOSECONELEVER:
		dg->ActivateDockingPort (my < 77 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;

	}
	return false;
}

DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	DeltaGlider *dg = (DeltaGlider*)vessel;

	if (KEYMOD_SHIFT (kstate)) {
	} else if (KEYMOD_CONTROL (kstate)) {
		switch (key) {
		case OAPI_KEY_B:
			dg->RevertAirbrake ();
			return 1;
		}
	} else {
		switch (key) {
		case OAPI_KEY_D:  // "operate radiator"
			dg->RevertRadiator ();
			return 1;
		case OAPI_KEY_G:  // "operate landing gear"
			dg->RevertLandingGear ();
			return 1;
		case OAPI_KEY_K:  // "operate docking port"
			dg->RevertDockingPort ();
			return 1;
		case OAPI_KEY_O:  // "operate outer airlock"
			dg->RevertAirlock ();
			return 1;
		}
	}
	return 0;
}
