// ==============================================================
//                ORBITER MODULE: DeltaGlider
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2003 Martin Schweiger
//                   All rights reserved
//
// DeltaGlider.h
// Class interface for delta glider vessel class module
//
// Notes:
// * This version uses the new thruster definition interface
// ==============================================================

#ifndef __DELTAGLIDER_H
#define __DELTAGLIDER_H

#include "orbitersdk.h"
#include "resource.h"

// ==========================================================
// Some vessel class caps
// ==========================================================

const double MAX_MAIN_FUEL = 11400.0;
// Max fuel capacity: main tank [kg]

const double MAX_RCS_FUEL = 600.0;
// Max fuel capacity: RCS tank [kg]

const double MAX_MAIN_THRUST = 1.2e5;
// Main engine max thust [N] per engine. Total thrust is twice this.

const double MAX_RETRO_THRUST = 2.7e4;
// Retro engine max thrust [N] per engine. Total thrust is twice this.

const double MAX_HOVER_THRUST = 2.7e5;
// Hover engine max thrust [N]

const double MAX_RCS_THRUST = 2.5e3;
// Attitude control system max thrust [N] per engine.

const double ISP = 4e4;
// Vacuum Isp (fuel-specific impulse) for all thrusters [m/s]

const double GEAR_OPERATING_SPEED = 0.15;
// Opening/closing speed of landing gear (1/sec)
// => gear cycle ~ 6.7 sec

const double NOSE_OPERATING_SPEED = 0.1;
// Opening/closing speed of nose cone docking mechanism (1/sec)
// => cycle = 10 sec

const double AIRLOCK_OPERATING_SPEED = 0.1;
// Opening/closing speed of outer airlock (1/sec)
// => cycle = 10 sec

const double RADIATOR_OPERATING_SPEED = 0.03125;
// Deployment speed of radiator (1/sec)
// => cycle = 32 sec

const double AIRBRAKE_OPERATING_SPEED = 0.4;
// Deployment speed of airbrakes
// => cycle = 2.5 sec

const int nsurf = 6; // number of bitmap handles

// ==========================================================
// Interface for derived vessel class: DeltaGlider
// ==========================================================

class DeltaGlider: public VESSEL {
public:
	DeltaGlider (OBJHANDLE hObj, int fmodel);
	~DeltaGlider ();
	void LoadState (FILEHANDLE scn, void *vs);
	void SaveState (FILEHANDLE scn);
	void SetClassCaps (FILEHANDLE cfg);
	void DefineAnimations ();
	void ReleaseSurfaces();
	void InitPanel (int panel);
	void Timestep (double simt);
	void DrawNeedle (HDC hDC, int x, int w, double rad, double angle, double *pangle = 0, double speed = PI);
	void RedrawPanel_PropellantStatus (SURFHANDLE surf);
	void RedrawPanel_EngineStatus (SURFHANDLE surf);
	void RedrawPanel_AOA (SURFHANDLE surf);
	void RedrawPanel_Slip (SURFHANDLE surf);
	void RedrawPanel_Slope (SURFHANDLE surf);
	void RedrawPanel_Range (SURFHANDLE surf);
	void RedrawPanel_Horizon (SURFHANDLE surf);
	void RedrawPanel_Navmode (SURFHANDLE surf);
	bool RedrawPanel_Throttle (SURFHANDLE surf);
	bool RedrawPanel_Hover (SURFHANDLE surf);
	void RedrawPanel_PGimbal (SURFHANDLE surf);
	void RedrawPanel_PYaw (SURFHANDLE surf);
	bool RedrawPanel_ElevTrim (SURFHANDLE surf);
	bool RedrawPanel_RotVelocity (SURFHANDLE surf, int which);
	bool RedrawPanel_RotAcceleration (SURFHANDLE surf, int which);
	bool RedrawPanel_RotMoment (SURFHANDLE surf, int which);
	bool RedrawPanel_WBrake (SURFHANDLE surf, int which);
	void RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side);
	bool DecAttMode ();
	bool IncAttMode ();
	bool DecADCMode ();
	bool IncADCMode ();
	void SetMainRetroLevel (int which, double lmain, double lretro);
	void SetGimbalPitch (int which, int pos);
	void SetGimbalYaw (int which, double tan_gmb);
	bool SwitchGimbalYawMode (bool switchup);
	bool SwitchGimbalYawComp (bool switchup);
	bool SwitchGimbalPitchMode (bool switchup);

	double alt;       // altitude
	double range;     // glide range
	double aoa_ind;   // angle of AOA needle (NOT AOA!)
	double slip_ind;  // angle of slip indicator needle
	double slope, slope_ind; // glide slope and angle of slope indicator needle
	int yg_mode, yg_comp, pg_mode;

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } nose_status, gear_status, lock_status, radiator_status, brake_status;
	void ActivateLandingGear (DoorStatus action);
	void ActivateDockingPort (DoorStatus action);
	void ActivateAirlock (DoorStatus action);
	void ActivateAirbrake (DoorStatus action);
	void ActivateRadiator (DoorStatus action);
	void RevertLandingGear ();
	void RevertDockingPort ();
	void RevertAirlock ();
	void RevertAirbrake ();
	void RevertRadiator ();
	void SetGearParameters (double state);
	double nose_proc, gear_proc, lock_proc, radiator_proc, brake_proc;     // logical status
	UINT anim_gear;         // handle for landing gear animation
	UINT anim_nose;         // handle for nose cone animation
	UINT anim_lock;         // handle for airlock animation
	UINT anim_radiator;     // handle for radiator animation
	UINT anim_rudder;       // handle for rudder animation
	UINT anim_elevator;     // handle for elevator animation
	UINT anim_elevatortrim; // handle for elevator trim animation
	UINT anim_laileron;     // handle for left aileron animation
	UINT anim_raileron;     // handle for right aileron animation
	UINT anim_brake;        // handle for airbrake animation
	SURFHANDLE srf[nsurf];  // handles for panel bitmaps
	THGROUP_HANDLE thg_main;
	THGROUP_HANDLE thg_retro;
	THGROUP_HANDLE thg_hover;

private:
	PROPELLANT_HANDLE ph_main, ph_rcs;
	THRUSTER_HANDLE th_main[2];
	THRUSTER_HANDLE th_retro[2];
	THRUSTER_HANDLE th_hover[1];
	UINT engsliderpos[3];    // throttle setting for left/right main engines
	UINT elevtrimpos;        // elevator trim indicator position
	int pg_pos[2], yg_pos[2]; // pitch and yaw gimbal slider positions
	double pgimbal[2], ygimbal[2]; // current gimbal angles
	UINT wbrake_pos[2];
};

typedef struct {
	HINSTANCE hDLL;
	HFONT font[2];
	DWORD col[4];
	HBRUSH brush[5];
	HPEN pen[2];
} GDIParams;

// Panel area identifiers:
// Panel 0
#define AID_PROPELLANTSTATUS     0
#define AID_ENGINESTATUS         1
#define AID_ENGINEMAIN           2
#define AID_ENGINEHOVER          3
#define AID_ATTITUDEMODE         4
#define AID_ADCTRLMODE           5
#define AID_NAVMODE              6
#define AID_HUDMODE              7
#define AID_AOAINSTR             8
#define AID_SLIPINSTR            9
#define AID_SLOPEINSTR          10
#define AID_RANGEINSTR          11
#define AID_HORIZON             12
#define AID_MFD1_BBUTTONS       13
#define AID_MFD1_LBUTTONS       14
#define AID_MFD1_RBUTTONS       15
#define AID_MFD2_BBUTTONS       16
#define AID_MFD2_LBUTTONS       17
#define AID_MFD2_RBUTTONS       18
#define AID_PITCHGIMBAL         19
#define AID_YAWGIMBAL           20
#define AID_YAWGIMBALMODE       21
#define AID_YAWGIMBALCOMP       22
#define AID_PITCHGIMBALMODE     23
#define AID_ELEVATORTRIM        24

// Panel 1
#define AID_OUTERAIRLOCKBUTTON 100
#define AID_VPITCH             101
#define AID_VBANK              102
#define AID_VYAW               103
#define AID_APITCH             104
#define AID_ABANK              105
#define AID_AYAW               106
#define AID_MPITCH             107
#define AID_MBANK              108
#define AID_MYAW               109

// Panel 2
#define AID_WBRAKE_BOTH        200
#define AID_WBRAKE_LEFT        201
#define AID_WBRAKE_RIGHT       202
#define AID_GEARLEVER          203
#define AID_NOSECONELEVER      204
#define AID_GEARINDICATOR      205
#define AID_NOSECONEINDICATOR  206

#endif // !__DELTAGLIDER_H