// ==============================================================
//                 ORBITER MODULE: Framerate
//                  Part of the ORBITER SDK
//            Copyright (C) 2003 Martin Schweiger
//                   All rights reserved
//
// Framerate.cpp
// Dialog box for displaying simulation frame rate.
// ==============================================================

#define STRICT
#define ORBITER_MODULE
#include <windows.h>
#include <stdio.h>
#include "Orbitersdk.h"
#include "Graph.h"
#include "resource.h"

static char *desc = "Simulation frame rate monitor";

// ==============================================================
// Global variables

HINSTANCE g_hInst;          // module instance handle
HWND g_hDlg;                // dialog handle
DWORD g_dwCmd;              // custom function identifier
Graph *g_Graph = 0;         // frame rate graph
double g_T = 0.0;           // sample time
double g_DT = 1.0;			// sample interval
DWORD g_fcount;             // frame counter

// ==============================================================
// Local prototypes

void OpenDlgClbk (void *context);
BOOL CALLBACK MsgProc (HWND, UINT, WPARAM, LPARAM);
long FAR PASCAL Framerate_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

// ==============================================================
// API interface

DLLCLBK void opcDLLInit (HINSTANCE hDLL)
{
	g_hInst = hDLL;
	g_hDlg = 0;
	g_dwCmd = oapiRegisterCustomCmd ("Frame Rate", desc, OpenDlgClbk, NULL);

	// register the window class for the data graph
	WNDCLASS wndClass;
	wndClass.style         = CS_HREDRAW | CS_VREDRAW;
	wndClass.lpfnWndProc   = Framerate_WndProc;
	wndClass.cbClsExtra    = 0;
	wndClass.cbWndExtra    = 0;
	wndClass.hInstance     = hDLL;
	wndClass.hIcon         = NULL;
	wndClass.hCursor       = LoadCursor (NULL, IDC_CROSS);
	wndClass.hbrBackground = (HBRUSH)GetStockObject (WHITE_BRUSH);
	wndClass.lpszMenuName  = NULL;
	wndClass.lpszClassName = "FramerateWindow";
	RegisterClass (&wndClass);

	Graph::InitGDI();
}

DLLCLBK void opcDLLExit (HINSTANCE hDLL)
{
	UnregisterClass ("FramerateWindow", g_hInst);
	oapiUnregisterCustomCmd (g_dwCmd);

	Graph::FreeGDI();
}

DLLCLBK void opcTimestep (double simdt, double mjd)
{
	if (!g_Graph) return; // flight data dialog not open

	double syst = oapiGetSysTime(); // ignore time acceleration for graph updates
	g_fcount++;

	if (syst >= g_T+g_DT) {
		char cbuf[128];
		float fps = (float)(g_fcount / (syst-g_T));
		sprintf (cbuf, "FPS: %0.0f", fps);
		g_Graph->AppendDataPoint (fps);
		g_Graph->SetTitle (cbuf);
		InvalidateRect (GetDlgItem (g_hDlg, IDC_FRAMERATE), NULL, TRUE);
		g_T = syst;
		g_fcount = 0;
	}
}

void OpenDlgClbk (void *context)
{
	HWND hDlg = oapiOpenDialog (g_hInst, IDD_FRAMERATE, MsgProc);
	if (hDlg) g_hDlg = hDlg; // otherwise open already
}

// =================================================================================
// FlightData dialog message handler
// =================================================================================

BOOL CALLBACK MsgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
	case WM_INITDIALOG:
		g_T = oapiGetSysTime();
		g_fcount = 0;
		g_Graph = new Graph(1);
		g_Graph->SetYLabel ("FPS");
		return TRUE;
	case WM_DESTROY:
		delete g_Graph;
		g_Graph = 0;
		return 0;
	case WM_SIZE: {
		RECT r;
		GetClientRect (hDlg, &r);
		SetWindowPos (GetDlgItem (hDlg, IDC_FRAMERATE), 0, 0, 0, r.right, r.bottom, SWP_NOMOVE | SWP_NOREPOSITION | SWP_NOZORDER);
		} return 0;
	case WM_COMMAND:
		switch (LOWORD (wParam)) {
		case IDCANCEL:
			oapiCloseDialog (g_hDlg);
			return TRUE;
		}
	}
	return oapiDefDialogProc (hDlg, uMsg, wParam, lParam);
}

// =================================================================================
// Graph canvas message handler
// =================================================================================

long FAR PASCAL Framerate_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
	case WM_PAINT: {
		RECT r;
		PAINTSTRUCT ps;
		HDC hDC;
		int gw, gh;
		GetClientRect (hWnd, &r);
		gw = r.right-r.left;
		gh = r.bottom-r.top;
		hDC = BeginPaint (hWnd, &ps);
		SetViewportOrgEx (hDC, 0, 0, NULL);
		g_Graph->Refresh (hDC, gw, gh);
		EndPaint (hWnd, &ps);
		} break;
	}
	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}

