// ==============================================================
//                 ORBITER MODULE:  ShuttleA
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2003 Martin Schweiger
//                   All rights reserved
//
// ShuttleA.cpp
// Reference implementation of Shuttle-A vessel class module
// ==============================================================

#define STRICT

#include "ShuttleA.h"
#include "resource.h"
#include <math.h>
#include <stdio.h>

#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))

// ==============================================================
// Global parameters
// ==============================================================

GDIParams g_Param;

// ==============================================================
// Specialised vessel class ShuttleA
// ==============================================================

// Constructor
ShuttleA::ShuttleA (OBJHANDLE hObj, int fmodel)
: VESSEL (hObj, fmodel)
{
	int i;
	dock_proc = 0.0;
	dock_status = DOOR_CLOSED;
	lock_proc = 0.0;
	lock_status = DOOR_CLOSED;
	DefineAnimations ();
	for (i = 0; i < nsurf; i++)
		srf[i] = 0;
	for (i = 0; i < 2; i++) {
		pod_angle[i] = pod_angle_request[i] = 0.0;
	}
}

// Destructor
ShuttleA::~ShuttleA ()
{
	ReleaseSurfaces();
}

// Set vessel class caps
void ShuttleA::SetClassCaps (FILEHANDLE cfg)
{
	int i;

	SetSize (17.0);
	SetPMI (_V(86.6, 89.8, 5.5));
	SetEmptyMass (EMPTY_MASS);
	SetCW (0.2, 0.2, 1.5, 1.5);
	SetCrossSections (_V(132.2, 237.9, 42.4));
	SetRotDrag (_V(0.7, 0.7, 0.3));
	SetSurfaceFrictionCoeff (0.5, 0.5);
	SetCameraOffset (_V(0,2.3,16));
	SetDockParams (_V(0,0,18.32), _V(0,0,1), _V(0,1,0));
	SetTouchdownPoints (_V(0,-2.93,9), _V(-2,-2.93,-8), _V(2,-2.93,-8));
	EnableTransponder (true);

	// **************************** NAV radios *************************************

	InitNavRadios (2);

	// ************************* propellant specs **********************************

	ph_main = CreatePropellantResource (MAX_MAIN_FUEL); // main tank
	ph_rcs  = CreatePropellantResource (MAX_RCS_FUEL);  // RCS tank

	// *********************** thruster definitions ********************************

	static PARTICLESTREAMSPEC contrail_main = {
		0, 8.0, 5, 150, 0.3, 4.0, 4, 2.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	static PARTICLESTREAMSPEC contrail_hover = {
		0, 6.0, 3, 150, 0.3, 4.0, 4, 2.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	static PARTICLESTREAMSPEC contrail_pod = {
		0, 4.0, 3, 150, 0.3, 3.0, 4, 2.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	static PARTICLESTREAMSPEC exhaust_main = {
		0, 2.0, 20, 150, 0.1, 0.2, 16, 2.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};

	// main thrusters
	th_main[0] = CreateThruster (_V(-1.8,0,-16.55), _V(0,0,1), MAX_MAIN_THRUST, ph_main, ISP);
	th_main[1] = CreateThruster (_V( 1.8,0,-16.55), _V(0,0,1), MAX_MAIN_THRUST, ph_main, ISP);
	thg_main = CreateThrusterGroup (th_main, 2, THGROUP_MAIN);
	AddExhaust (th_main[0], 12, 1);
	AddExhaust (th_main[1], 12, 1);
	AddExhaustStream (th_main[0], _V(0,0,-24), &contrail_main);
	AddExhaustStream (th_main[0], _V(-1.8,0,-19), &exhaust_main);
	AddExhaustStream (th_main[1], _V( 1.8,0,-19), &exhaust_main);

	// hover and retro thrusters
	th_hover[0] = CreateThruster (_V(0,-2.4, 13.25), _V(0,1,0), MAX_HOVER_THRUST, ph_main, ISP);
	th_hover[1] = CreateThruster (_V(0,-2.4,-13.25), _V(0,1,0), MAX_HOVER_THRUST, ph_main, ISP);
	thg_hover = CreateThrusterGroup (th_hover, 2, THGROUP_HOVER);
	AddExhaust (th_hover[0], 10, 1);
	AddExhaust (th_hover[1], 10, 1);
	AddExhaustStream (th_hover[0], _V(0,-6, 13.25), &contrail_hover);
	AddExhaustStream (th_hover[1], _V(0,-6,-13.25), &contrail_hover);
	AddExhaustStream (th_hover[0], _V(0,-4, 13.25), &exhaust_main);
	AddExhaustStream (th_hover[1], _V(0,-4,-13.25), &exhaust_main);

	// retro/hover thrusters
	th_pod[0] = CreateThruster (_V(-7,0,0/*1.5*/), _V(0,0,-1), MAX_RETRO_THRUST, ph_main, ISP);
	th_pod[1] = CreateThruster (_V( 7,0,0/*1.5*/), _V(0,0,-1), MAX_RETRO_THRUST, ph_main, ISP);
	thg_pod = CreateThrusterGroup (th_pod, 2, THGROUP_USER);
	AddExhaust (th_pod[0], 6, 0.7, 1.5);
	AddExhaust (th_pod[1], 6, 0.7, 1.5);
	AddExhaustStream (th_pod[0], &contrail_pod);
	AddExhaustStream (th_pod[1], &contrail_pod);

	// attitude thrusters
	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];
	th_att_rot[0] = CreateThruster (_V(-6, 0.7,-0.4), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = CreateThruster (_V( 6,-0.7,-0.4), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = CreateThruster (_V(-6,-0.7,-0.4), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = CreateThruster (_V( 6, 0.7,-0.4), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);
	for (i = 0; i < 4; i++) AddExhaust (th_att_rot[i], 0.7, 0.08);
	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = th_att_lin[2] = CreateThruster (_V(0,0,-15), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = th_att_lin[3] = CreateThruster (_V(0,0, 15), _V(0,-1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15), _V(0, 1,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_UP);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_DOWN);
	AddExhaust (th_att_rot[0], 0.7, 0.08, _V(-0.2, -1.1,   17.5), _V(0,-1,0));
	AddExhaust (th_att_rot[0], 0.7, 0.08, _V( 0.2, -1.1,   17.5), _V(0,-1,0));
	AddExhaust (th_att_rot[1], 0.7, 0.08, _V( 0,    2.3,-14.6), _V(0, 1,0));
	AddExhaust (th_att_rot[1], 0.7, 0.08, _V( 0,    2.3,-14.3), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.7, 0.08, _V(-0.15, 2.1, 17.6), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.7, 0.08, _V( 0.15, 2.1, 17.6), _V(0, 1,0));
	AddExhaust (th_att_rot[3], 0.7, 0.08, _V(-0.15,-1.9,-14.5), _V(0,-1,0));
	AddExhaust (th_att_rot[3], 0.7, 0.08, _V( 0.15,-1.9,-14.5), _V(0,-1,0));
	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15), _V( 1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[1] = th_att_lin[2] = CreateThruster (_V(0,0,-15), _V(-1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[2] = th_att_lin[3] = CreateThruster (_V(0,0, 15), _V(-1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15), _V( 1,0,0), MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWRIGHT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_RIGHT);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_LEFT);
	AddExhaust (th_att_rot[0], 0.7, 0.08, _V(-2.4, 0.15, 17.5), _V(-1,0,0));
	AddExhaust (th_att_rot[0], 0.7, 0.08, _V(-2.4,-0.15, 17.5), _V(-1,0,0));
	AddExhaust (th_att_rot[1], 0.7, 0.08, _V( 3.5, 0,   -14.6), _V( 1,0,0));
	AddExhaust (th_att_rot[1], 0.7, 0.08, _V( 3.5, 0,   -14.3), _V( 1,0,0));
	AddExhaust (th_att_rot[2], 0.7, 0.08, _V( 2.4, 0.15, 17.5), _V( 1,0,0));
	AddExhaust (th_att_rot[2], 0.7, 0.08, _V( 2.4,-0.15, 17.5), _V( 1,0,0));
	AddExhaust (th_att_rot[3], 0.7, 0.08, _V(-3.5, 0,   -14.6), _V(-1,0,0));
	AddExhaust (th_att_rot[3], 0.7, 0.08, _V(-3.5, 0,   -14.3), _V(-1,0,0));
	th_att_lin[0] = CreateThruster (_V( 0,0,0), _V(0,0, 1), 2*MAX_RCS_THRUST, ph_rcs, ISP);
	th_att_lin[1] = CreateThruster (_V( 0,0,0), _V(0,0,-1), 2*MAX_RCS_THRUST, ph_rcs, ISP);
	CreateThrusterGroup (th_att_lin,   1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_att_lin+1, 1, THGROUP_ATT_BACK);
	AddExhaust (th_att_lin[0], 0.7, 0.08, _V( 6,0,-1.6), _V(0,0,-1));
	AddExhaust (th_att_lin[0], 0.7, 0.08, _V(-6,0,-1.6), _V(0,0,-1));
	AddExhaust (th_att_lin[1], 0.7, 0.08, _V( 6,0, 1), _V(0,0, 1));
	AddExhaust (th_att_lin[1], 0.7, 0.08, _V(-6,0, 1), _V(0,0, 1));

	// ******************************** mesh ***************************************

	AddMesh (oapiLoadMeshGlobal ("ShuttleA"));
}

// Read status from scenario file
void ShuttleA::LoadState (FILEHANDLE scn, void *vs)
{
	char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "PODANGLE", 8)) {
			sscanf (line+8, "%lf%lf", pod_angle+0, pod_angle+1);
		} else if (!strnicmp (line, "DOCKSTATE", 9)) {
			sscanf (line+9, "%d%lf", &dock_status, &dock_proc);
		} else if (!strnicmp (line, "AIRLOCK", 7)) {
			sscanf (line+7, "%d%lf", &lock_status, &lock_proc);
		} else {
			ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
		}
	}
}

// Write status to scenario file
void ShuttleA::SaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	sprintf (cbuf, "%0.4f %0.4f", pod_angle[0], pod_angle[1]);
	oapiWriteScenario_string (scn, "PODANGLE", cbuf);

	sprintf (cbuf, "%d %0.4f", dock_status, dock_proc);
	oapiWriteScenario_string (scn, "DOCKSTATE", cbuf);

	sprintf (cbuf, "%d %0.4f", lock_status, lock_proc);
	oapiWriteScenario_string (scn, "AIRLOCK", cbuf);
}

void ShuttleA::PostCreation ()
{
	for (UINT i = 0; i < 2; i++) {
		pod_angle_request[i] = pod_angle[i];
		double sina = sin(pod_angle[i]), cosa = cos(pod_angle[i]);
		SetAnimState (anim_pod[i], pod_angle[i]/PI);
		SetThrusterDir (th_pod[i], _V(0,sina,-cosa));
	}
	SetAnimState (anim_dock, dock_proc);
	SetAnimState (anim_lock, lock_proc);
}

// Define animation sequences for moving parts
void ShuttleA::DefineAnimations ()
{
	static UINT LeftPodGrp[4] = {62,63,64,65};
	static ANIMCOMP leftpod = {
		LeftPodGrp, 4, 0, 1,
		0, 0, 0, //-0.35,
		1, 0, 0,
		(float)PI,
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT RightPodGrp[4] = {66,67,68,69};
	static ANIMCOMP rightpod = {
		RightPodGrp, 4, 0, 1,
		0, 0, 0, //-0.35,
		1, 0, 0,
		(float)PI,
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT UpperDockHatch = 48;
	static ANIMCOMP upperhatch = {
		&UpperDockHatch, 1, 0, 1,
		0, 0.554, 18.317,
		-1, 0, 0,
		(float)PI,
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT LowerDockHatch = 49;
	static ANIMCOMP lowerhatch = {
		&LowerDockHatch, 1, 0, 1,
		0, -0.554, 18.317,
		1, 0, 0,
		(float)PI,
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT OuterAirlock = 21;
	static ANIMCOMP outerairlock = {
		&OuterAirlock, 1, 0, 1,
		0, 0.495, 18.252,
		1, 0, 0,
		(float)(0.4*PI),
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};

	// Register animation for hover/retro pods
	anim_pod[0] = RegisterAnimSequence (0.0);
	AddAnimComp (anim_pod[0], &leftpod);
	anim_pod[1] = RegisterAnimSequence (0.0);
	AddAnimComp (anim_pod[1], &rightpod);

	anim_dock = RegisterAnimSequence (0.0);
	AddAnimComp (anim_dock, &upperhatch);
	AddAnimComp (anim_dock, &lowerhatch);

	anim_lock = RegisterAnimSequence (0.0);
	AddAnimComp (anim_lock, &outerairlock);
}

// Frame update
void ShuttleA::Timestep (double simt)
{
	// animate auxiliary engine pods
	bool redraw = false;
	for (int i = 0; i < 2; i++) {
		if (pod_angle[i] != pod_angle_request[i]) {
			double da = oapiGetSimStep() * POD_ROT_SPEED;
			if (pod_angle[i] < pod_angle_request[i]) {
				pod_angle[i] += da;
				if (pod_angle[i] > pod_angle_request[i])
					pod_angle[i] = pod_angle_request[i];
			} else {
				pod_angle[i] -= da;
				if (pod_angle[i] < pod_angle_request[i])
					pod_angle[i] = pod_angle_request[i];
			}
			double sina = sin(pod_angle[i]), cosa = cos(pod_angle[i]);
			SetThrusterDir (th_pod[i], _V(0,sina,-cosa));
			SetAnimState (anim_pod[i], pod_angle[i]/PI);
			redraw = true;
		}
	}
	if (redraw) oapiTriggerPanelRedrawArea (0, AID_PODANGLEINDICATOR);

	// animate docking hatch
	if (dock_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * DOCK_OPERATING_SPEED;
		if (dock_status == DOOR_CLOSING) {
			if (dock_proc > 0.0)
				dock_proc = max (0.0, dock_proc-da);
			else {
				dock_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
			}
		} else { // hatch opening
			if (dock_proc < 1.0)
				dock_proc = min (1.0, dock_proc+da);
			else {
				dock_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
			}
		}
		SetAnimState (anim_dock, dock_proc);
	}

	// animate airlock
	if (lock_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * AIRLOCK_OPERATING_SPEED;
		if (lock_status == DOOR_CLOSING) {
			if (lock_proc > 0.0)
				lock_proc = max (0.0, lock_proc-da);
			else {
				lock_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (1, AID_AIRLOCK1INDICATOR);
			}
		} else { // door opening
			if (lock_proc < 1.0)
				lock_proc = min (1.0, lock_proc+da);
			else {
				lock_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (1, AID_AIRLOCK1INDICATOR);
			}
		}
		SetAnimState (anim_lock, lock_proc);
	}
}

void ShuttleA::ReleaseSurfaces ()
{
	for (int i = 0; i < nsurf; i++)
		if (srf[i]) {
			oapiDestroySurface (srf[i]);
			srf[i] = 0;
		}
}

void ShuttleA::InitPanel (int panel)
{
	int i;

	switch (panel) {
	case 0:
		srf[0] = oapiCreateSurface (LOADBMP (IDB_SLIDER1));
		srf[1] = oapiCreateSurface (LOADBMP (IDB_SWITCH1));
		srf[2] = oapiCreateSurface (LOADBMP (IDB_BUTTON1));
		for (i = 0; i < 2; i++) {
			sliderpos_main[i] = sliderpos_hovr[i] =
			sliderpos_retro[i] = sliderpos_auxhovr[i] = 
			sliderpos_pod[i] = (UINT)-1;
			podswitch[i] = 0;
		}
		break;
	case 1:
		srf[0] = oapiCreateSurface (LOADBMP (IDB_SWITCH2));
		srf[1] = oapiCreateSurface (LOADBMP (IDB_SWITCH3));
		srf[2] = oapiCreateSurface (LOADBMP (IDB_SWITCH4));
		srf[3] = oapiCreateSurface (LOADBMP (IDB_INDICATOR1));
		srf[4] = oapiCreateSurface (LOADBMP (IDB_INDICATOR2));
		break;
	}
}

void ShuttleA::RotatePod (UINT which, UINT mode)
{
	if (!mode) return; // nothing to do

	double dt = oapiGetSimStep();

	if (mode == 2) { // turn retro
		pod_angle_request[which] -= dt*POD_ROTREQUEST_SPEED;
		if (pod_angle_request[which] < 0.0) pod_angle_request[which] = 0.0;
	} else {         // turn forward
		pod_angle_request[which] += dt*POD_ROTREQUEST_SPEED;
		if (pod_angle_request[which] > PI) pod_angle_request[which] = PI;
	}
	oapiTriggerPanelRedrawArea (0, AID_PODANGLEINDICATOR);
}

void ShuttleA::PresetPods (UINT mode)
{
	double preset_angle[3] = {PI, 0.5*PI, 0};
	for (int i = 0; i < 2; i++)
		pod_angle_request[i] = (2-mode)*0.5*PI;
	oapiTriggerPanelRedrawArea (0, AID_PODANGLEINDICATOR);
}

void ShuttleA::ActivateDockingPort (DoorStatus action)
{
	dock_status = action;
	oapiTriggerPanelRedrawArea (1, AID_DOCKSWITCH);
	oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
}

void ShuttleA::RevertDockingPort ()
{
	ActivateDockingPort (dock_status == DOOR_CLOSED || dock_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void ShuttleA::ActivateAirlock (DoorStatus action)
{
	lock_status = action;
	oapiTriggerPanelRedrawArea (1, AID_AIRLOCK1SWITCH);
	oapiTriggerPanelRedrawArea (1, AID_AIRLOCK1INDICATOR);
}

void ShuttleA::RevertAirlock ()
{
	ActivateAirlock (lock_status == DOOR_CLOSED || lock_status == DOOR_CLOSING ?
		             DOOR_OPENING : DOOR_CLOSING);
}

void ShuttleA::RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side)
{
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.hFont[0]);
	SetTextColor (hDC, RGB(0, 200, 0));
	SetTextAlign (hDC, TA_CENTER);
	SetBkMode (hDC, TRANSPARENT);
	const char *label;
	for (int bt = 0; bt < 6; bt++) {
		if (label = oapiMFDButtonLabel (mfd, bt+side*6))
			TextOut (hDC, 13, 3+38*bt, label, strlen(label));
		else break;
	}
	oapiReleaseDC (surf, hDC);
}

void ShuttleA::RedrawPanel_Navmode (SURFHANDLE surf)
{
	for (DWORD i = NAVMODE_KILLROT; i < NAVMODE_HOLDALT; i++)
		if (GetNavmodeState (i))
			oapiBlt (surf, srf[2], (6-i)*44, 0, (i-1)*42, 0, 42, 31);
}

bool ShuttleA::RedrawPanel_Throttle (SURFHANDLE surf)
{
	UINT i, pos;
	bool redraw = false;

	for (i = 0; i < 2; i++) {
		double level = GetThrusterLevel (th_main[i]);
		pos = (UINT)((1.0-level)*180.0);
		if (pos != sliderpos_main[i])
			sliderpos_main[i] = pos, redraw = true;
	}
	if (redraw)
		for (i = 0; i < 2; i++)
			oapiBlt (surf, srf[0], i*30, sliderpos_main[i], 0, 0, 23, 19);
	return redraw;
}

bool ShuttleA::RedrawPanel_Hover (SURFHANDLE surf)
{
	UINT i, pos;
	bool redraw = false;

	for (i = 0; i < 2; i++) {
		double level = GetThrusterLevel (th_hover[i]);
		pos = (UINT)((1.0-level)*180.0);
		if (pos != sliderpos_hovr[i])
			sliderpos_hovr[i] = pos, redraw = true;
	}
	if (redraw)
		for (i = 0; i < 2; i++)
			oapiBlt (surf, srf[0], i*30, sliderpos_hovr[i], 0, 0, 23, 19);
	return redraw;
}

bool ShuttleA::RedrawPanel_Podlevel (SURFHANDLE surf)
{
	UINT i, pos;
	bool redraw = false;

	for (i = 0; i < 2; i++) {
		double level = GetThrusterLevel (th_pod[i]);
		pos = (UINT)((1.0-level)*90.0);
		if (pos != sliderpos_pod[i])
			sliderpos_pod[i] = pos, redraw = true;
	}
	if (redraw)
		for (i = 0; i < 2; i++)
			oapiBlt (surf, srf[0], i*30, sliderpos_pod[i], 0, 0, 23, 19);
	return redraw;
}

bool ShuttleA::RedrawPanel_EngineIndicator (SURFHANDLE surf)
{
	const double rad = 25.0;
	const int cntx[2] = {29, 95};
	const int txtx[2] = {16, 83};
	const int cnty[3] = {29, 96, 163};
	const int txty0a =  34, txty0b =  46;
	const int txty1a = 101, txty1b = 113;
	const int txty2a = 168, txty2b = 180;

	UINT i;
	double level, th, angle, dx, dy;
	char cbuf[16];
	double m = GetMass();

	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.hFont[0]);
	SelectObject (hDC, g_Param.hPen[0]);
	SetTextColor (hDC, RGB(120,220,120));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);

	for (i = 0; i < 2; i++) {
		level = GetThrusterLevel (th_main[i]);
		th    = level*MAX_MAIN_THRUST;
		angle = level * (1.5*PI);
		dx = rad*cos(angle), dy = rad*sin(angle);
		MoveToEx (hDC, cntx[i], cnty[0], NULL);
		LineTo (hDC, cntx[i]-(int)dx, cnty[0]-(int)dy);
		sprintf (cbuf, "%0.0f", th*1e-3);
		TextOut (hDC, txtx[i], txty0a, cbuf, strlen(cbuf));
		sprintf (cbuf, "%0.1f", th/m);
		TextOut (hDC, txtx[i], txty0b, cbuf, strlen (cbuf));

		level = GetThrusterLevel (th_hover[i]);
		th    = level*MAX_HOVER_THRUST;
		angle = level * (1.5*PI);
		dx = rad*cos(angle), dy = rad*sin(angle);
		MoveToEx (hDC, cntx[i], cnty[1], NULL);
		LineTo (hDC, cntx[i]-(int)dx, cnty[1]-(int)dy);
		sprintf (cbuf, "%0.0f", th*1e-3);
		TextOut (hDC, txtx[i], txty1a, cbuf, strlen(cbuf));
		sprintf (cbuf, "%0.1f", th/m);
		TextOut (hDC, txtx[i], txty1b, cbuf, strlen (cbuf));

		level = GetThrusterLevel (th_pod[i]);
		th    = level*MAX_RETRO_THRUST;
		angle = level * (1.5*PI);
		dx = rad*cos(angle), dy = rad*sin(angle);
		MoveToEx (hDC, cntx[i], cnty[2], NULL);
		LineTo (hDC, cntx[i]-(int)dx, cnty[2]-(int)dy);
		sprintf (cbuf, "%0.0f", th*1e-3);
		TextOut (hDC, txtx[i], txty2a, cbuf, strlen(cbuf));
		sprintf (cbuf, "%0.1f", th/m);
		TextOut (hDC, txtx[i], txty2b, cbuf, strlen (cbuf));
	}
	oapiReleaseDC (surf, hDC);

	return true;
}

bool ShuttleA::RedrawPanel_PodangleIndicator (SURFHANDLE surf)
{
	const int cntx[2] = {36,110}, cnty = 14;
	const int txtx[2] = {45,119}, txty =  0;
	const double rad = 24.0, radi = 29.0, radw = 36.0;
	const double da = 0.1;

	UINT i;
	int x, y, ia;
	double angle, angle1, angle2;
	char cbuf[16];

	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.hFont[0]);
	SetBkMode (hDC, TRANSPARENT);

	for (i = 0; i < 2; i++) {
		// draw preset indicator
		SelectObject (hDC, g_Param.hPen[1]);
		angle = pod_angle_request[i];
		angle1 = angle-da;
		angle2 = angle+da;
		x = cntx[i]-(int)(radi*cos(angle)), y = cnty+(int)(radi*sin(angle));
		MoveToEx (hDC, x, y, NULL);
		LineTo (hDC, cntx[i]-(int)(radw*cos(angle1)), cnty+(int)(radw*sin(angle1)));
		LineTo (hDC, cntx[i]-(int)(radw*cos(angle2)), cnty+(int)(radw*sin(angle2)));
		LineTo (hDC, x, y);
		ia = (int)(DEG*angle+0.5);
		if      (ia < 90) sprintf (cbuf, "%02dR", ia);
		else if (ia > 90) sprintf (cbuf, "%02dF", 180-ia);
		else              sprintf (cbuf, "90");
		SetTextColor (hDC, RGB(220,220,120));
		TextOut (hDC, txtx[i]-27, txty, cbuf, strlen(cbuf));

		// draw pod status indicator
		SelectObject (hDC, g_Param.hPen[0]);
		angle = pod_angle[i];
		MoveToEx (hDC, cntx[i], cnty, NULL);
		LineTo (hDC, cntx[i]-(int)(rad*cos(angle)), cnty+(int)(rad*sin(angle)));
		ia = (int)(DEG*angle+0.5);
		if      (ia < 90) sprintf (cbuf, "%02dR", ia);
		else if (ia > 90) sprintf (cbuf, "%02dF", 180-ia);
		else              sprintf (cbuf, "90");
		SetTextColor (hDC, RGB(120,220,120));
		TextOut (hDC, txtx[i], txty, cbuf, strlen(cbuf));
	}
	oapiReleaseDC (surf, hDC);
	return true;
}

void ShuttleA::RedrawPanel_Fuelstatus (SURFHANDLE surf, int part)
{
	char cbuf[20];
	double m, m0, rate, lvl;

	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.hFont[0]);
	SetTextColor (hDC, RGB(224,224,224));
	SetBkMode (hDC, TRANSPARENT);

	switch (part) {
	case 0:
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_hover[0]) * MAX_HOVER_THRUST / ISP);
		TextOut (hDC, 0, 0, cbuf, strlen (cbuf));
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_hover[1]) * MAX_HOVER_THRUST / ISP);
		TextOut (hDC, 0, 30, cbuf, strlen (cbuf));
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_pod[1]) * MAX_RETRO_THRUST / ISP);
		TextOut (hDC, 0, 58, cbuf, strlen (cbuf));
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_pod[0]) * MAX_RETRO_THRUST / ISP);
		TextOut (hDC, 0, 88, cbuf, strlen (cbuf));
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_main[0]) * MAX_MAIN_THRUST / ISP);
		TextOut (hDC, 0, 115, cbuf, strlen (cbuf));
		sprintf (cbuf, "%0.1f", GetThrusterLevel (th_main[1]) * MAX_MAIN_THRUST / ISP);
		TextOut (hDC, 0, 145, cbuf, 3);
		break;
	case 1:
		m = GetPropellantMass (ph_main);
		if (m > MAX_MAIN_FUEL*0.2) {
			rate = GetPropellantFlowrate (ph_main);
			lvl = m*1.25/MAX_MAIN_FUEL - 0.25;
			SelectObject (hDC, g_Param.hBrush[0]);
			Rectangle (hDC,  0, 50, 28, (int)((1.0-lvl)*50.0));
			Rectangle (hDC, 38, 50, 66, (int)((1.0-lvl)*50.0));
			m0 = 0.5 * (m - MAX_MAIN_FUEL*0.2);
		} else {
			rate = lvl = m0 = 0;
		}
		sprintf (cbuf, "%0.1f", 0.5*rate);
		TextOut (hDC, 19, 55, cbuf, strlen(cbuf));
		TextOut (hDC, 57, 55, cbuf, strlen(cbuf));
		sprintf (cbuf, "%0.0f", m0);
		SetTextAlign (hDC, TA_CENTER);
		TextOut (hDC, 14, 20, cbuf, strlen (cbuf));
		TextOut (hDC, 52, 20, cbuf, strlen (cbuf));
		break;
	case 2:
		m = GetPropellantMass (ph_main);
		if (m < MAX_MAIN_FUEL*0.2) {
			rate = GetPropellantFlowrate (ph_main);
			lvl = 5.0*m/MAX_MAIN_FUEL;
		} else {
			rate = 0;
			lvl = 1;
			m = MAX_MAIN_FUEL*0.2;
		}
		if (lvl > 0) {
			SelectObject (hDC, g_Param.hBrush[0]);
			Rectangle (hDC,  0, 57, 28, 19+(int)((1.0-lvl)*38.0));
		}
		sprintf (cbuf, "%0.1f", rate);
		TextOut (hDC, 19, 0, cbuf, strlen(cbuf));
		sprintf (cbuf, "%0.0f", m);
		SetTextAlign (hDC, TA_CENTER);
		TextOut (hDC, 14, 33, cbuf, strlen (cbuf));
		break;
	case 3:
		m = GetPropellantMass (ph_rcs);
		if (m > 0) {
			SelectObject (hDC, g_Param.hBrush[0]);
			Rectangle (hDC, 0, 25, 28, (int)((1.0-m/MAX_RCS_FUEL)*25.0));
		}
		sprintf (cbuf, "%0.2f", GetPropellantFlowrate (ph_rcs));
		TextOut (hDC, 19, 28, cbuf, strlen(cbuf));
		SetTextAlign (hDC, TA_CENTER);
		sprintf (cbuf, "%0.1f", m);
		TextOut (hDC, 14, 7, cbuf, strlen(cbuf));
		break;
	} 
	oapiReleaseDC (surf, hDC);
}

// ==============================================================
// DLL entry point
// ==============================================================

BOOL WINAPI DllMain (HINSTANCE hModule,
					 DWORD ul_reason_for_call,
					 LPVOID lpReserved)
{
	int i;

	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		g_Param.hDLL = hModule;
		g_Param.hFont[0] = CreateFont (-10, 0, 0, 0, 400, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
		g_Param.hPen[0] = CreatePen (PS_SOLID, 3, RGB (120,220,120));
		g_Param.hPen[1] = CreatePen (PS_SOLID, 1, RGB (220,220,120));
		g_Param.hBrush[0] = CreateSolidBrush (RGB(0,128,0));
		break;
	case DLL_PROCESS_DETACH:
		for (i = 0; i < 1; i++) DeleteObject (g_Param.hFont[i]);
		for (i = 0; i < 2; i++) DeleteObject (g_Param.hPen[i]);
		for (i = 0; i < 1; i++) DeleteObject (g_Param.hBrush[i]);
		break;
	}
	return TRUE;
}

// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new ShuttleA (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (ShuttleA*)vessel;
}

// Set vessel class caps
DLLCLBK void ovcSetClassCaps (VESSEL* vessel, FILEHANDLE cfg)
{
	((ShuttleA*)vessel)->SetClassCaps (cfg);
}

// Read status from scenario file
DLLCLBK void ovcLoadStateEx (VESSEL *vessel, FILEHANDLE scn, void *vs)
{
	((ShuttleA*)vessel)->LoadState (scn, vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((ShuttleA*)vessel)->SaveState (scn);
}

DLLCLBK void ovcPostCreation (VESSEL *vessel)
{
	((ShuttleA*)vessel)->PostCreation();

}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((ShuttleA*)vessel)->Timestep (simt);
}

DLLCLBK bool ovcLoadPanel (VESSEL *vessel, int id)
{
	static MFDSPEC mfds_left  = {{ 66, 84,  326, 344}, 6, 6, 31, 38};
	static MFDSPEC mfds_right = {{828, 84, 1088, 344}, 6, 6, 31, 38};
	ShuttleA *sh = (ShuttleA*)vessel;
	sh->ReleaseSurfaces();

	HBITMAP hBmp = LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (IDB_PANEL1+id));

	switch (id) {
	case 0: // main panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_BOTTOM|PANEL_MOVEOUT_BOTTOM, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,1,-1);
		oapiRegisterMFD (MFD_LEFT, mfds_left);
		oapiRegisterMFD (MFD_RIGHT, mfds_right);
		oapiRegisterPanelArea (AID_ENGINEMAIN,        _R( 414,  43, 467, 242), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINEHOVER,       _R( 504,  43, 557, 242), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINEPODLEVEL,    _R( 414, 250, 467, 359), PANEL_REDRAW_MOUSE, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINEINDICATOR,   _R( 584,  47, 709, 237), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_PODANGLEINDICATOR, _R( 502, 308, 649, 358), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_PODANGLESWITCH,    _R( 521, 263, 571, 307), PANEL_REDRAW_MOUSE, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_LBUP, PANEL_MAP_CURRENT);
		oapiRegisterPanelArea (AID_PODANGLEPRESET,    _R( 605, 253, 647, 303), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_NAVMODE,           _R( 112,  17, 374,  48), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ATTITUDEMODE,      _R( 719,  62, 739, 106), PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiRegisterPanelArea (AID_MFD1_LBUTTONS,     _R(  28, 109,  52, 318), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_MFD1_RBUTTONS,     _R( 338, 109, 362, 318), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_MFD1_BBUTTONS,     _R( 129, 349, 262, 368), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_MFD2_LBUTTONS,     _R( 790, 109, 814, 318), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_MFD2_RBUTTONS,     _R(1100, 109,1124, 318), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_MFD2_BBUTTONS,     _R( 891, 349,1024, 368), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN);
		break;
	case 1: // overhead panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_TOP|PANEL_MOVEOUT_TOP, 0xffffff);
		oapiRegisterPanelArea (AID_FUELSTATUS1,       _R( 191, 102, 207, 257), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_FUELSTATUS2,       _R( 233,  44, 310, 109), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_FUELSTATUS3,       _R( 266, 211, 305, 267), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_FUELSTATUS4,       _R( 182,  44, 221,  82), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_AIRLOCK1SWITCH,    _R( 477,  38, 501,  88), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_AIRLOCK1INDICATOR, _R( 473,  91, 507,  99), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_DOCKSWITCH,        _R( 479, 128, 503, 178), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_DOCKINDICATOR,     _R( 475, 180, 509, 188), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiSetPanelNeighbours (-1,-1,-1,0);
		break;
	}
	sh->InitPanel (id);
	return hBmp != NULL;
}

DLLCLBK bool ovcPanelRedrawEvent (VESSEL *vessel, int id, int event, SURFHANDLE surf)
{
	ShuttleA *sh = (ShuttleA*)vessel;

	switch (id) {
	// panel 0 events:
	case AID_MFD1_LBUTTONS:
		sh->RedrawPanel_MFDButton (surf, MFD_LEFT, 0);
		return true;
	case AID_MFD1_RBUTTONS:
		sh->RedrawPanel_MFDButton (surf, MFD_LEFT, 1);
		return true;
	case AID_MFD2_LBUTTONS:
		sh->RedrawPanel_MFDButton (surf, MFD_RIGHT, 0);
		return true;
	case AID_MFD2_RBUTTONS:
		sh->RedrawPanel_MFDButton (surf, MFD_RIGHT, 1);
		return true;
	case AID_ENGINEMAIN:
		return sh->RedrawPanel_Throttle (surf);
	case AID_ENGINEHOVER:
		return sh->RedrawPanel_Hover (surf);
	case AID_ENGINEPODLEVEL:
		return sh->RedrawPanel_Podlevel (surf);
	case AID_ENGINEINDICATOR:
		return sh->RedrawPanel_EngineIndicator (surf);
	case AID_PODANGLEINDICATOR:
		return sh->RedrawPanel_PodangleIndicator (surf);
	case AID_PODANGLESWITCH:
		oapiBlt (surf, sh->srf[1],  0, 0, sh->podswitch[0]*20, 0, 20, 44);
		oapiBlt (surf, sh->srf[1], 30, 0, sh->podswitch[1]*20, 0, 20, 44);
		return true;
	case AID_NAVMODE:
		sh->RedrawPanel_Navmode (surf);
		return true;
	case AID_ATTITUDEMODE:
		oapiBlt (surf, sh->srf[1], 0, 0, ((sh->GetAttitudeMode()+2)%3)*20, 0, 20, 44);
		return true;

	// panel 1 events
	case AID_FUELSTATUS1:
	case AID_FUELSTATUS2:
	case AID_FUELSTATUS3:
	case AID_FUELSTATUS4:
		sh->RedrawPanel_Fuelstatus (surf, id-AID_FUELSTATUS1);
		return true;
	case AID_AIRLOCK1SWITCH:
		oapiBlt (surf, sh->srf[0], 0, 0, sh->lock_status == ShuttleA::DOOR_OPEN ||
			sh->lock_status == ShuttleA::DOOR_OPENING ? 24:0, 0, 24, 50);
		return true;
	case AID_AIRLOCK1INDICATOR:
		switch (sh->lock_status) {
		case ShuttleA::DOOR_CLOSED: oapiBlt (surf, sh->srf[3], 0, 0, 0,  0, 34, 8); break;
		case ShuttleA::DOOR_OPEN:   oapiBlt (surf, sh->srf[3], 0, 0, 0, 16, 34, 8); break;
		default:                    oapiBlt (surf, sh->srf[3], 0, 0, 0,  8, 34, 8); break;
		}
		return true;
	case AID_DOCKSWITCH:
		oapiBlt (surf, sh->srf[2], 0, 0, sh->dock_status == ShuttleA::DOOR_OPEN ||
			sh->dock_status == ShuttleA::DOOR_OPENING ? 24:0, 0, 24, 50);
		return true;
	case AID_DOCKINDICATOR:
		switch (sh->dock_status) {
		case ShuttleA::DOOR_CLOSED: oapiBlt (surf, sh->srf[4], 0, 0, 0,  0, 34, 8); break;
		case ShuttleA::DOOR_OPEN:   oapiBlt (surf, sh->srf[4], 0, 0, 0, 16, 34, 8); break;
		default:                    oapiBlt (surf, sh->srf[4], 0, 0, 0,  8, 34, 8); break;
		}
		return true;
	}
	return false;
}

DLLCLBK bool ovcPanelMouseEvent (VESSEL *vessel, int id, int event, int mx, int my)
{
	ShuttleA *sh = (ShuttleA*)vessel;
	static int ctrl = 3;
	double lvl;
	int mode, pmode, mfd;

	switch (id) {
	// panel 0 events:
	case AID_MFD1_LBUTTONS:
	case AID_MFD1_RBUTTONS:
	case AID_MFD2_LBUTTONS:
	case AID_MFD2_RBUTTONS:
		if (my%39 < 20) {
			mfd = (id <= AID_MFD1_RBUTTONS ? MFD_LEFT : MFD_RIGHT);
			int bt = my/39 + (id == AID_MFD1_LBUTTONS || id == AID_MFD2_LBUTTONS ? 0 : 6);
			oapiProcessMFDButton (mfd, bt, event);
			return true;
		}
		break;
	case AID_MFD1_BBUTTONS:
	case AID_MFD2_BBUTTONS:
		mfd = (id == AID_MFD1_BBUTTONS ? MFD_LEFT : MFD_RIGHT);
		if      (mx <  32)            oapiToggleMFD_on (mfd);
		else if (mx >= 50 && mx < 82) oapiSendMFDKey (mfd, OAPI_KEY_F1);
		else if (mx >= 100)           oapiSendMFDKey (mfd, OAPI_KEY_GRAVE);
		return true;
	case AID_ENGINEMAIN:
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (mx <  12) ctrl = 1;  // left engine
			else if (mx >= 41) ctrl = 2;  // right engine
			else               ctrl = 3;  // both
		}
		if ((my -= 9) < 0) lvl = 1.0;
		else if (my > 180) lvl = 0.0;
		else               lvl = 1.0-my/180.0;
		if (ctrl & 1) sh->SetThrusterLevel (sh->th_main[0], lvl);
		if (ctrl & 2) sh->SetThrusterLevel (sh->th_main[1], lvl);
		return true;
	case AID_ENGINEHOVER:
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (mx <  12) ctrl = 1;  // front engine
			else if (mx >= 41) ctrl = 2;  // back engine
			else               ctrl = 3;  // both
		}
		if ((my -= 9) < 0) lvl = 1.0;
		else if (my > 180) lvl = 0.0;
		else               lvl = 1.0-my/180.0;
		if (ctrl & 1) sh->SetThrusterLevel (sh->th_hover[0], lvl);
		if (ctrl & 2) sh->SetThrusterLevel (sh->th_hover[1], lvl);
		return true;
	case AID_ENGINEPODLEVEL:
		if (event & PANEL_MOUSE_LBDOWN) { // record which slider to operate
			if      (mx <  12) ctrl = 1;  // left engine
			else if (mx >= 41) ctrl = 2;  // right engine
			else               ctrl = 3;  // both
		}
		if ((my -= 9) < 0) lvl = 1.0;
		else if (my > 90)  lvl = 0.0;
		else               lvl = 1.0-my/90.0;
		if (ctrl & 1) sh->SetThrusterLevel (sh->th_pod[0], lvl);
		if (ctrl & 2) sh->SetThrusterLevel (sh->th_pod[1], lvl);
		return true;
	case AID_PODANGLESWITCH:
		if (event & PANEL_MOUSE_LBDOWN) { // record which switch to operate
			if      (mx <  10) ctrl = 1;  // left pod
			else if (mx >= 38) ctrl = 2;  // right pod
			else               ctrl = 3;  // both
		}
		pmode = ((ctrl & 1) ? sh->podswitch[0] : sh->podswitch[1]);
		if (event & PANEL_MOUSE_LBUP) mode = 0;
		else                          mode = (my < 22 ? 1:2);
		if (ctrl & 1) sh->RotatePod (0, sh->podswitch[0] = mode);
		if (ctrl & 2) sh->RotatePod (1, sh->podswitch[1] = mode);
		return mode != pmode;
	case AID_PODANGLEPRESET:
		if ((my % 18) > 14) return false;
		mode = my / 18;
		sh->PresetPods (mode);
		return true;
	case AID_NAVMODE:
		if ((mx % 44) < 42) {
			sh->ToggleNavmode (6 - mx/44);
			return true;
		}
		break;
	case AID_ATTITUDEMODE:
		pmode = sh->GetAttitudeMode();
		if (my < 22) {
			static int nmode[3] = {1,2,2};
			mode = nmode[pmode];
		} else {
			static int nmode[3] = {0,0,1};
			mode = nmode[pmode];
		}
		if (mode != pmode) sh->SetAttitudeMode (mode);
		return mode != pmode;

	// panel 1 events
	case AID_AIRLOCK1SWITCH:
		sh->ActivateAirlock (my < 25 ? ShuttleA::DOOR_CLOSING : ShuttleA::DOOR_OPENING);
		break;
	case AID_DOCKSWITCH:
		sh->ActivateDockingPort (my < 25 ? ShuttleA::DOOR_CLOSING : ShuttleA::DOOR_OPENING);
		break;
	}
	return false;
}

DLLCLBK void ovcMFDmode (VESSEL *vessel, int mfd, int mode)
{
	switch (mfd) {
	case MFD_LEFT:
		oapiTriggerPanelRedrawArea (0, AID_MFD1_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD1_RBUTTONS);
		break;
	case MFD_RIGHT:
		oapiTriggerPanelRedrawArea (0, AID_MFD2_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD2_RBUTTONS);
		break;
	}
}

DLLCLBK void ovcNavmode (VESSEL *vessel, int mode, bool active)
{
	oapiTriggerPanelRedrawArea (0, AID_NAVMODE);
}

DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	ShuttleA *sh = (ShuttleA*)vessel;

	if (KEYMOD_SHIFT (kstate)) {
	} else if (KEYMOD_CONTROL (kstate)) {
		switch (key) {
		case OAPI_KEY_DIVIDE:
			if (vessel->SetAttitudeMode (vessel->GetAttitudeMode() >= 1 ? 0 : 1))
				oapiTriggerPanelRedrawArea (0, AID_ATTITUDEMODE);
			return 1;
		}
	} else {
		switch (key) {
		case OAPI_KEY_K:  // "operate docking port"
			sh->RevertDockingPort ();
			return 1;
		case OAPI_KEY_O:  // "operate outer airlock"
			sh->RevertAirlock ();
			return 1;
		case OAPI_KEY_DIVIDE:
			if (vessel->ToggleAttitudeMode()) oapiTriggerPanelRedrawArea (0, AID_ATTITUDEMODE);
			return 1;
		}
	}
	return 0;
}
