// ==============================================================
//                 ORBITER MODULE:  ShuttleA
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2003 Martin Schweiger
//                   All rights reserved
//
// ShuttleA.h
// Class interface for Shuttle-A vessel class module
// ==============================================================

#ifndef __SHUTTLEA_H
#define __SHUTTLEA_H

#include "orbitersdk.h"

// ==========================================================
// Some vessel class caps
// ==========================================================

const double EMPTY_MASS = 22000.0;
// Empty mass (excluding fuel) [kg]

const double MAX_MAIN_FUEL = 16000.0;
// Max fuel capacity: main tank [kg]

const double MAX_RCS_FUEL = 700.0;
// Max fuel capacity: RCS tank [kg]

const double MAX_MAIN_THRUST = 193520.0;
// Main engine thrust [N] per engine

const double MAX_HOVER_THRUST = 135450.0;
// Hover engine thrust [N] per engine

const double MAX_RETRO_THRUST = 60000.0;
// Max thrust [N] of retro/hover pods per engine

const double MAX_RCS_THRUST = 2500.0;
// Max thrust [N] for each attitude thruster

const double ISP = 3e4;
// Vacuum Isp (fuel-specific impulse) for all thrusters [m/s]

const double DOCK_OPERATING_SPEED = 0.05;
// Opening/closing speed of docking hatch mechanism (1/sec)
// => cycle = 20 sec

const double AIRLOCK_OPERATING_SPEED = 0.1;
// Opening/closing speed of outer airlock (1/sec)
// => cycle = 10 sec

const double POD_ROTREQUEST_SPEED = 0.4;
// speed at which the pod tilt switches react to preset changes

const double POD_ROT_SPEED = 0.2;
// speed at which pods actually rotate

const int nsurf = 5; // number of bitmap handles

// ==========================================================
// Interface for derived vessel class: ShuttleA
// ==========================================================

class ShuttleA: public VESSEL {
public:
	ShuttleA (OBJHANDLE hObj, int fmodel);
	~ShuttleA ();
	void SetClassCaps (FILEHANDLE cfg);
	void LoadState (FILEHANDLE scn, void *vs);
	void SaveState (FILEHANDLE scn);
	void PostCreation ();
	void Timestep (double simt);
	void ReleaseSurfaces ();
	void InitPanel (int panel);
	void RotatePod (UINT which, UINT mode);
	void PresetPods (UINT mode);

	void RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side);
	void RedrawPanel_Navmode (SURFHANDLE surf);
	bool RedrawPanel_Throttle (SURFHANDLE surf);
	bool RedrawPanel_Hover (SURFHANDLE surf);
	bool RedrawPanel_Podlevel (SURFHANDLE surf);
	bool RedrawPanel_EngineIndicator (SURFHANDLE surf);
	bool RedrawPanel_PodangleIndicator (SURFHANDLE surf);
	void RedrawPanel_Fuelstatus (SURFHANDLE surf, int part);

	SURFHANDLE srf[nsurf];
	PROPELLANT_HANDLE ph_main, ph_rcs;
	THRUSTER_HANDLE th_main[2], th_hover[2], th_pod[2];
	UINT podswitch[2];
	double dock_proc, lock_proc;
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } dock_status, lock_status;

	void ActivateDockingPort (DoorStatus action);
	void RevertDockingPort ();
	void ActivateAirlock (DoorStatus action);
	void RevertAirlock ();

private:
	void DefineAnimations ();
	THGROUP_HANDLE thg_main, thg_hover, thg_pod;
	double pod_angle[2], pod_angle_request[2];
	UINT anim_pod[2], anim_dock, anim_lock;
	UINT sliderpos_main[2], sliderpos_hovr[2];
	UINT sliderpos_retro[2], sliderpos_auxhovr[2], sliderpos_pod[2];
};

typedef struct {
	HINSTANCE hDLL;
	HFONT hFont[1];
	HPEN hPen[2];
	HBRUSH hBrush[1];
} GDIParams;

#define AID_MFD1_LBUTTONS      0
#define AID_MFD1_RBUTTONS      1
#define AID_MFD1_BBUTTONS      2
#define AID_MFD2_LBUTTONS      3
#define AID_MFD2_RBUTTONS      4
#define AID_MFD2_BBUTTONS      5

#define AID_ENGINEMAIN         6
#define AID_ENGINEHOVER        7
#define AID_ENGINEPODLEVEL     8
#define AID_ENGINEINDICATOR    9
#define AID_PODANGLEINDICATOR 10
#define AID_PODANGLESWITCH    11
#define AID_PODANGLEPRESET    12
#define AID_NAVMODE           13
#define AID_ATTITUDEMODE      14

#define AID_FUELSTATUS1       20
#define AID_FUELSTATUS2       21
#define AID_FUELSTATUS3       22
#define AID_FUELSTATUS4       23
#define AID_AIRLOCK1SWITCH    24
#define AID_AIRLOCK1INDICATOR 25
#define AID_DOCKSWITCH        26
#define AID_DOCKINDICATOR     27

#endif // !__SHUTTLEA_H