// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2003 Martin Schweiger
//                   All rights reserved
//
// HST.cpp
// HST basic specs and animations
//
// HST mesh and textures by David Sundstrom
// ==============================================================

#define STRICT
#include "HST.h"
#include <stdio.h>

// ==============================================================
// HST class implementation
// ==============================================================

HST::HST (OBJHANDLE hObj, int fmodel)
: VESSEL (hObj, fmodel)
{
	ant_proc = 0.0;
	ant_status = DOOR_CLOSED;
	hatch_proc = 0.0;
	hatch_status = DOOR_CLOSED;
	fold_proc = 1.0;
	fold_status = DOOR_OPEN;
	DefineAnimations ();
}

void HST::SetClassCaps (FILEHANDLE cfg)
{
	SetSize (HST_SIZE);
	SetEmptyMass (HST_EMPTY_MASS);
}

void HST::DefineAnimations (void)
{
	// 1. Hi-gain antenna
	static UINT HiGainAnt1Grp[2] = {1,3};
	static MGROUP_ROTATE HiGainAnt1 (0, HiGainAnt1Grp, 2, _V(0.002579,1.993670,0.238158), _V(-1,0,0), (float)(PI*0.51));
	static UINT HiGainAnt2Grp[2] = {0,2};
	static MGROUP_ROTATE HiGainAnt2 (0, HiGainAnt2Grp, 2, _V(0.002740,-2.013091,0.238118), _V(1,0,0), (float)(PI*0.51));
	anim_ant = CreateAnimation (0.0196);
	AddAnimationComponent (anim_ant, 0, 0.5, &HiGainAnt1);
	AddAnimationComponent (anim_ant, 0, 1,   &HiGainAnt2);

	// 2. Main telescope hatch
	static UINT HatchGrp[1] = {86};
	static MGROUP_ROTATE Hatch (0, HatchGrp, 1, _V(0.089688,1.456229,7.526453), _V(-1,0,0), (float)(RAD*113));
	anim_hatch = CreateAnimation (0);
	AddAnimationComponent (anim_hatch, 0, 1, &Hatch);

	// 3. Solar arrays - folding
	anim_fold = CreateAnimation (1);
	static UINT ArrayLFoldGrp[5] = {87,88,89,90,103};
	static UINT ArrayRFoldGrp[5] = {92,93,94,95,102};
	static MGROUP_ROTATE ArrayLFold1 (0, ArrayLFoldGrp, 5, _V(-1.9, 0.053583,1.429349), _V(0,-1,0), (float)(PI*0.5));
	AddAnimationComponent (anim_fold, 0,   0.4, &ArrayLFold1);
	static MGROUP_ROTATE ArrayLFold2 (0, ArrayLFoldGrp, 5, _V(0,0.053583,1.429349), _V(-1,0,0), (float)(PI*0.5));
	AddAnimationComponent (anim_fold, 0.4, 0.6, &ArrayLFold2);
	static MGROUP_SCALE  ArrayLFold3 (0, ArrayLFoldGrp, 4, _V(0,0.053583,1.429349), _V(1,1,4));
	AddAnimationComponent (anim_fold, 0.6, 1,   &ArrayLFold3);
	static MGROUP_ROTATE ArrayRFold1 (0, ArrayRFoldGrp, 5, _V( 1.9, 0.053583,1.429349), _V(0, 1,0), (float)(PI*0.5));
	AddAnimationComponent (anim_fold, 0,   0.4, &ArrayRFold1);
	static MGROUP_ROTATE ArrayRFold2 (0, ArrayRFoldGrp, 5, _V(0,0.053583,1.429349), _V(-1,0,0), (float)(PI*0.5));
	AddAnimationComponent (anim_fold, 0.4, 0.6, &ArrayRFold2);
	static MGROUP_SCALE  ArrayRFold3 (0, ArrayRFoldGrp, 4, _V(0,0.053583,1.429349), _V(1,1,4));
	AddAnimationComponent (anim_fold, 0.6, 1,   &ArrayRFold3);
}

void HST::LoadState (FILEHANDLE scn, void *vs)
{
	char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ANT", 3)) {
			sscanf (line+3, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "HATCH", 5)) {
			sscanf (line+5, "%d%lf", &hatch_status, &hatch_proc);
		} else if (!strnicmp (line, "FOLD", 4)) {
			sscanf (line+5, "%d%lf", &fold_status, &fold_proc);
		} else {
			ParseScenarioLineEx (line, vs);
		}
	}

	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_hatch, hatch_proc);
	SetAnimation (anim_fold, fold_proc);
}

void HST::SaveState (FILEHANDLE scn)
{
	char cbuf[256];
	SaveDefaultState (scn);
	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "ANT", cbuf);
	sprintf (cbuf, "%d %0.4f", hatch_status, hatch_proc);
	oapiWriteScenario_string (scn, "HATCH", cbuf);
	sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	oapiWriteScenario_string (scn, "FOLD", cbuf);
}

void HST::Timestep (double simt)
{
	// Animate hi-gain antenna
	if (ant_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * ANTENNA_OPERATING_SPEED;
		if (ant_status == DOOR_CLOSING) {
			if (ant_proc > 0.0) ant_proc = max (0.0, ant_proc-da);
			else                ant_status = DOOR_CLOSED;
		} else {
			if (ant_proc < 1.0) ant_proc = min (1.0, ant_proc+da);
			else                ant_status = DOOR_OPEN;
		}
		SetAnimation (anim_ant, ant_proc);
	}

	// Animate main telescope hatch
	if (hatch_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * HATCH_OPERATING_SPEED;
		if (hatch_status == DOOR_CLOSING) {
			if (hatch_proc > 0.0) hatch_proc = max (0.0, hatch_proc-da);
			else                  hatch_status = DOOR_CLOSED;
		} else {
			if (hatch_proc < 1.0) hatch_proc = min (1.0, hatch_proc+da);
			else                  hatch_status = DOOR_OPEN;
		}
		SetAnimation (anim_hatch, hatch_proc);
	}

	// Animate solar arrays
	if (fold_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * ARRAY_OPERATING_SPEED;
		if (fold_status == DOOR_CLOSING) {
			if (fold_proc > 0.0) fold_proc = max (0.0, fold_proc-da);
			else                 fold_status = DOOR_CLOSED;
		} else {
			if (fold_proc < 1.0) fold_proc = min (1.0, fold_proc+da);
			else                 fold_status = DOOR_OPEN;
		}
		SetAnimation (anim_fold, fold_proc);
	}
}

void HST::RevertAntenna (void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void HST::RevertHatch (void)
{
	hatch_status = ((hatch_status == DOOR_CLOSED || hatch_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void HST::RevertArray (void)
{
	fold_status = ((fold_status == DOOR_CLOSED || fold_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// API interface
// ==============================================================

// Vessel initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new HST (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (HST*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	((HST*)vessel)->SetClassCaps (cfg);
}

// Read status from scenario file
DLLCLBK void ovcLoadStateEx (VESSEL *vessel, FILEHANDLE scn, void *vs)
{
	((HST*)vessel)->LoadState (scn, vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((HST*)vessel)->SaveState (scn);
}

// Frame update
DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((HST*)vessel)->Timestep (simt);
}

// Keyboard interface handler (buffered key events)
DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	HST *hst = (HST*)vessel;

	if (KEYMOD_CONTROL (kstate)) {

		switch (key) {
		case OAPI_KEY_1: // deploy/retract antenna
			hst->RevertAntenna();
			return 1;
		case OAPI_KEY_2: // open/close hatch
			hst->RevertHatch();
			return 1;
		case OAPI_KEY_3: // open/fold solar arrays
			hst->RevertArray();
			return 1;
		}
	}
	return 0;
}
